import math
from functools import partial
from typing import Union, Dict, List, Tuple, Optional, Optional

import numpy as np
import pandas as pd

from PyQt5 import QtCore, QtWidgets
import pyqtgraph as pg

# from matplotlib.backends import backend_qt5agg
import matplotlib as mpl
import matplotlib.pyplot as plt

from sanpy.sanpyLogger import get_logger

logger = get_logger(__name__)

import sanpy
from sanpy.interface.plugins import sanpyPlugin


class spikeClips(sanpyPlugin):
    """Plot AP clips."""

    myHumanName = "Plot Spike Clips"

    def __init__(self, **kwargs):
        """
        Args:
            ba (bAnalysis): Not required
        """
        super().__init__(**kwargs)

        self.preClipWidth_ms = 100
        self.postClipWidth_ms = 500
        if self.ba is not None and self.ba.isAnalyzed():
            # self.clipWidth_ms = self.ba.detectionClass['spikeClipWidth_ms']
            self.preClipWidth_ms = self.ba._detectionDict["preSpikeClipWidth_ms"]
            self.postClipWidth_ms = self.ba._detectionDict["postSpikeClipWidth_ms"]

        self.respondTo = "All"  # ('All', 'X-Axis', 'Spike Selection')

        # keep track so we can export to clipboard
        self.x = None
        self.y = None
        self.yMean = None

        # for waterfall
        self.xMult = 0.1
        self.yMult = 0.1

        # self.clipLines = None
        # self.meanClipLine = None
        self.singleSpikeMultiLine = None
        self.spikeListMultiLine = None

        # main layout
        vLayout = QtWidgets.QVBoxLayout()

        #
        # controls
        hLayout2 = QtWidgets.QHBoxLayout()

        self.numSpikesLabel = QtWidgets.QLabel("Num Spikes:None")
        hLayout2.addWidget(self.numSpikesLabel)

        self.colorCheckBox = QtWidgets.QCheckBox("Color")
        self.colorCheckBox.setChecked(False)
        self.colorCheckBox.stateChanged.connect(lambda: self.replot())
        hLayout2.addWidget(self.colorCheckBox)

        self.meanCheckBox = QtWidgets.QCheckBox("Mean Trace (red)")
        self.meanCheckBox.setChecked(True)
        self.meanCheckBox.stateChanged.connect(lambda: self.replot())
        hLayout2.addWidget(self.meanCheckBox)

        self.varianceCheckBox = QtWidgets.QCheckBox("Variance")
        self.varianceCheckBox.setChecked(False)
        self.varianceCheckBox.stateChanged.connect(lambda: self.replot())
        hLayout2.addWidget(self.varianceCheckBox)

        self.phasePlotCheckBox = QtWidgets.QCheckBox("Phase")
        self.phasePlotCheckBox.setChecked(False)
        self.phasePlotCheckBox.stateChanged.connect(lambda: self.replot())
        hLayout2.addWidget(self.phasePlotCheckBox)

        vLayout.addLayout(hLayout2)

        #
        hLayout2_5 = QtWidgets.QHBoxLayout()

        self.waterfallCheckBox = QtWidgets.QCheckBox("Waterfall")
        self.waterfallCheckBox.setChecked(False)
        self.waterfallCheckBox.stateChanged.connect(lambda: self.replot())
        hLayout2_5.addWidget(self.waterfallCheckBox)

        # x mult for waterfall
        aLabel = QtWidgets.QLabel("X-Offset")
        hLayout2_5.addWidget(aLabel)
        self.xMultSpinBox = QtWidgets.QDoubleSpinBox()
        self.xMultSpinBox.setKeyboardTracking(False)
        self.xMultSpinBox.setSingleStep(0.05)
        self.xMultSpinBox.setRange(0, 1)
        self.xMultSpinBox.setValue(self.xMult)
        self.xMultSpinBox.valueChanged.connect(partial(self._on_spinbox, aLabel))
        hLayout2_5.addWidget(self.xMultSpinBox)

        # y mult for waterfall
        aLabel = QtWidgets.QLabel("Y-Offset")
        hLayout2_5.addWidget(aLabel)
        self.yMultSpinBox = QtWidgets.QDoubleSpinBox()
        self.yMultSpinBox.setKeyboardTracking(False)
        self.yMultSpinBox.setSingleStep(0.05)
        self.yMultSpinBox.setRange(0, 1)
        self.yMultSpinBox.setValue(self.yMult)
        self.yMultSpinBox.valueChanged.connect(partial(self._on_spinbox, aLabel))
        hLayout2_5.addWidget(self.yMultSpinBox)

        vLayout.addLayout(hLayout2_5)

        #
        hLayout3 = QtWidgets.QHBoxLayout()

        aLabel = QtWidgets.QLabel("Pre (ms)")
        hLayout3.addWidget(aLabel)
        self.preClipWidthSpinBox = QtWidgets.QSpinBox()
        self.preClipWidthSpinBox.setKeyboardTracking(False)
        self.preClipWidthSpinBox.setRange(1, 2**16)
        self.preClipWidthSpinBox.setValue(self.preClipWidth_ms)
        # self.clipWidthSpinBox.editingFinished.connect(partial(self.on_spinbox, aLabel))
        self.preClipWidthSpinBox.valueChanged.connect(partial(self._on_spinbox, aLabel))
        hLayout3.addWidget(self.preClipWidthSpinBox)

        aLabel = QtWidgets.QLabel("Post (ms)")
        hLayout3.addWidget(aLabel)
        self.postClipWidthSpinBox = QtWidgets.QSpinBox()
        self.postClipWidthSpinBox.setKeyboardTracking(False)
        self.postClipWidthSpinBox.setRange(1, 2**16)
        self.postClipWidthSpinBox.setValue(self.postClipWidth_ms)
        # self.clipWidthSpinBox.editingFinished.connect(partial(self.on_spinbox, aLabel))
        self.postClipWidthSpinBox.valueChanged.connect(partial(self._on_spinbox, aLabel))
        hLayout3.addWidget(self.postClipWidthSpinBox)

        #
        # radio buttons
        self.respondToAll = QtWidgets.QRadioButton("All")
        self.respondToAll.setChecked(True)
        self.respondToAll.toggled.connect(self.on_radio)
        hLayout3.addWidget(self.respondToAll)

        self.respondToAxisRange = QtWidgets.QRadioButton("X-Axis")
        self.respondToAxisRange.setChecked(False)
        self.respondToAxisRange.toggled.connect(self.on_radio)
        hLayout3.addWidget(self.respondToAxisRange)

        self.respondToSpikeSel = QtWidgets.QRadioButton("Spike Selection")
        self.respondToSpikeSel.setChecked(False)
        self.respondToSpikeSel.toggled.connect(self.on_radio)
        hLayout3.addWidget(self.respondToSpikeSel)

        #
        vLayout.addLayout(hLayout3)

        #
        # pyqt graph
        self.view = pg.GraphicsLayoutWidget()
        self.clipPlot = self.view.addPlot(row=0, col=0)

        self.clipPlot.enableAutoRange()

        # self.clipPlot.hideButtons()
        # self.clipPlot.setMenuEnabled(False)
        # self.clipPlot.setMouseEnabled(x=False, y=False)

        self.clipPlot.getAxis("left").setLabel("mV")
        self.clipPlot.getAxis("bottom").setLabel("time (ms)")

        # kymograph analysis
        self.variancePlot = self.view.addPlot(row=1, col=0)

        vLayout.addWidget(self.view)  # , stretch=8)

        # set the layout of the main window
        # self.mainWidget.setLayout(vLayout)
        # self.setLayout(vLayout)
        self.getVBoxLayout().addLayout(vLayout)

        self.replot()

    def slot_switchFile(
        self, ba: sanpy.bAnalysis, rowDict: Optional[dict] = None, replot: bool = True
    ):
        # logger.info('')

        # don't replot until we set our detectionClass
        replot = False
        super().slot_switchFile(ba, rowDict, replot=replot)

        if self.ba is not None and self.ba.isAnalyzed():
            self.preClipWidth_ms = self.ba.getDetectionDict()["preSpikeClipWidth_ms"]
            self.postClipWidth_ms = self.ba.getDetectionDict()["postSpikeClipWidth_ms"]

        self.replot()

    def on_radio(self):
        """
        Will receive this callback n times where n is # buttons/checkboxes in group
        Only one callback will have isChecked, all other will be !isChecked
        """
        # logger.info('')
        sender = self.sender()
        senderText = sender.text()
        isChecked = sender.isChecked()

        if senderText == "All" and isChecked:
            self.respondTo = "All"
        elif senderText == "X-Axis" and isChecked:
            self.respondTo = "X-Axis"
        elif senderText == "Spike Selection" and isChecked:
            self.respondTo = "Spike Selection"
            # debug
            # self.selectedSpikeList = [5, 7, 10]

        #
        if isChecked:
            self.replot()

    def _on_spinbox(self, label):
        # logger.info('')
        # grab from interface
        self.preClipWidth_ms = self.preClipWidthSpinBox.value()
        self.postClipWidth_ms = self.postClipWidthSpinBox.value()
        self.xMult = self.xMultSpinBox.value()
        self.yMult = self.yMultSpinBox.value()

        self.replot()

    def setAxis(self):
        self.replot()

    def replot(self):
        """Replot when analysis changes or file changes"""
        self._myReplotClips()
        #
        # self.mainWidget.repaint() # update the widget

    '''
    def setAxis(self):
        """
        (self.startSec, self.stopSec) has been set by sanpyPlugins
        """
        self.replot()
    '''

    def _myReplotClips(self):
        """
        Note: This is the same code as in bDetectionWidget.refreshClips() MERGE THEM.
        """

        # logger.info('')

        isPhasePlot = self.phasePlotCheckBox.isChecked()

        # always remove existing
        self.clipPlot.clear()

        # self.clipPlot.enableAutoRange()

        self.variancePlot.clear()

        if self.ba is None or self.ba.numSpikes == 0:
            return

        # TODO: Add option to select by selectSpikeList

        #
        # respond to x-axis selection
        startSec = None
        stopSec = None
        selectedSpikeList = []
        logger.info(f'self.respondTo: {self.respondTo}')
        if self.respondTo == "All":
            startSec = 0
            stopSec = self.ba.fileLoader.recordingDur
        elif self.respondTo == "X-Axis":
            startSec, stopSec = self.getStartStop()
            if startSec is None or stopSec is None:
                startSec = 0
                stopSec = self.ba.fileLoader.recordingDur
        elif self.respondTo == "Spike Selection":
            selectedSpikeList = self.getSelectedSpikes()

        logger.info(f'  startSec:{startSec} stopSec:{stopSec}')

        # print('=== selectedSpikeList:', selectedSpikeList)

        # this returns x-axis in ms
        # theseClips is a [list] of clips
        # theseClips_x is in ms
        theseClips, theseClips_x, meanClip = self.ba.getSpikeClips(
            startSec,
            stopSec,
            spikeSelection=selectedSpikeList,
            preSpikeClipWidth_ms=self.preClipWidth_ms,
            postSpikeClipWidth_ms=self.postClipWidth_ms,
            sweepNumber=self.sweepNumber,
            epochNumber=self.epochNumber,
            ignoreMinMax=False  # 20230418 trying to get sweeps/epochs working
        )
        numClips = len(theseClips)
        logger.info(f'  got numClips:{numClips}')
        self.numSpikesLabel.setText(f"Num Spikes: {numClips}")

        if numClips == 0:
            return

        # convert clips to 2d ndarray ???
        xTmp = np.array(theseClips_x)
        xTmp /= 1000  # ms to seconds
        yTmp = np.array(theseClips)  # mV

        if isPhasePlot:
            # plot x mV versus y dV/dt
            dvdt = np.zeros((yTmp.shape[0], yTmp.shape[1] - 1))  #
            for i in range(dvdt.shape[0]):
                dvdt[i, :] = np.diff(yTmp[i, :])
                # drop first pnt in x

            #
            xTmp = yTmp[:, 1:]  # drop first column of mV and swap to x-axis
            yTmp = dvdt
            # print(xTmp.shape, yTmp.shape)

        # for waterfall we need x-axis to have different values for each spike
        if self.waterfallCheckBox.isChecked():
            # print(xTmp.shape, yTmp.shape)
            # xTmp and yTmp are 2D with rows/clips and cols/data_pnts
            xMin = np.nanmin(xTmp)
            xMax = np.nanmax(xTmp)
            xRange = xMax - xMin
            yMin = np.nanmin(yTmp)
            yMax = np.nanmax(yTmp)
            yRange = yMax - yMin
            xOffset = 0
            yOffset = 0
            xInc = xRange * self.xMult  # ms, xInc is 10% of x-range
            yInc = yRange * self.yMult
            for i in range(xTmp.shape[0]):
                xTmp[i, :] += xOffset
                yTmp[i, :] += yOffset
                xOffset += xInc  # ms
                yOffset += yInc  # mV

        #
        # original, one multiline for all clips (super fast)
        # self.clipLines = MultiLine(xTmp, yTmp, self, allowXAxisDrag=False, type='clip')
        #
        # each clip so we can set color
        # cmap = pg.colormap.get('Greens', source='matplotlib') # prepare a linear color map
        numSpikes = xTmp.shape[0]

        doColor = self.colorCheckBox.isChecked()
        if doColor:
            # cmap = pg.colormap.get('CET-L18') # prepare a linear color map
            cmap = pg.colormap.get(
                "gist_rainbow", source="matplotlib"
            )  # prepare a linear color map

            tmpColors = cmap.getColors()
            tmpLinSpace = np.linspace(
                0, len(tmpColors), numSpikes, endpoint=False, dtype=np.uint16
            )
        # for tmp in tmps:
        #    print(tmp)
        for i in range(numSpikes):
            # forcePenColor = cmap.getByIndex(i)  # returns PyQt5.QtGui.QColor
            forcePenColor = None
            if doColor:
                currentStep = tmpLinSpace[i]
                forcePenColor = tmpColors[currentStep]
            else:
                forcePenColor = self.getPenColor()

            xPlot = xTmp[i, :]
            yPlot = yTmp[i, :]
            tmpClipLines = MultiLine(
                xPlot,
                yPlot,
                self,
                allowXAxisDrag=False,
                forcePenColor=forcePenColor,
                type="clip",
            )
            self.clipPlot.addItem(tmpClipLines)

        #
        # always calculate mean clip
        # x mean is redundant but works well for waterfall
        xMeanClip = np.nanmean(xTmp, axis=0)  # xTmp is in ms
        yMeanClip = np.nanmean(yTmp, axis=0)

        # plot if checkbox is on
        if self.meanCheckBox.isChecked():
            tmpMeanClipLine = MultiLine(
                xMeanClip,
                yMeanClip,
                self,
                width=3,
                allowXAxisDrag=False,
                type="meanclip",
            )
            self.clipPlot.addItem(tmpMeanClipLine)

        #
        if not self.varianceCheckBox.isChecked():
            self.variancePlot.hide()
        else:
            self.variancePlot.show()
            # self.variancePlot.clear()
            xVarClip = np.nanmean(xTmp, axis=0)  # xTmp is in ms
            yVarClip = np.nanvar(yTmp, axis=0)
            tmpVarClipLine = MultiLine(
                xVarClip, yVarClip, self, width=3, allowXAxisDrag=False, type="meanclip"
            )
            self.variancePlot.addItem(tmpVarClipLine)
            self.variancePlot.getAxis("left").setLabel("Variance")

        # set axis
        xLabel = "time (s)"
        yLabel = self.ba.fileLoader.get_yUnits()
        if isPhasePlot:
            xLabel = self.ba.fileLoader.get_yUnits()
            yLabel = "d/dt"
        self.clipPlot.getAxis("left").setLabel(yLabel)
        self.clipPlot.getAxis("bottom").setLabel(xLabel)

        # mar 27 2023
        self.clipPlot.autoRange()  # 20230327

        #
        # store so we can export
        # print('  xTmp:', xTmp.shape)
        # print('  yTmp:', yTmp.shape)
        # print('  yMeanClip:', yMeanClip.shape)
        self.x = xTmp  # 1D
        self.y = yTmp  # 2D
        self.yMean = yMeanClip

        # update pre/post clip (ms)
        self.preClipWidthSpinBox.setValue(self.preClipWidth_ms)
        self.postClipWidthSpinBox.setValue(self.postClipWidth_ms)

        #
        # replot any selected spikes
        # self.old_selectSpike()
        self.selectSpikeList()

    def saveResultsFigure(self):
        super().saveResultsFigure(pgPlot=self.clipPlot)

    def copyToClipboard(self):
        """
        Save instead
        Copy to clipboard is not going to work, data is too complex/big
        """

        numSpikes = self.y.shape[0]

        columns = ["time(ms)", "meanClip"]
        # iterate through spikes
        for i in range(numSpikes):
            columns.append(f"clip_{i}")
        df = pd.DataFrame(columns=columns)

        # print('self.x:', type(self.x), self.x.shape)

        df["time(ms)"] = self.x[0, :]
        df["meanClip"] = self.yMean
        for i in range(numSpikes):
            df[f"clip_{i}"] = self.y[i]

        super().copyToClipboard(df=df)

    def old_selectSpike(self, sDict=None):
        """
        Leave existing spikes and select one spike with a different color.
        The one spike might not be displayed, then do nothing.
        sDict (dict): NOT USED
        """

        if sDict is not None:
            spikeNumber = sDict["spikeNumber"]
        else:
            spikeNumber = self.selectedSpike

        x = np.zeros(1) * np.nan
        y = np.zeros(1) * np.nan
        if spikeNumber is not None:
            try:
                x = self.x[spikeNumber]
                y = self.y[spikeNumber]
            except IndexError as e:
                pass

        # TODO: Fix this. Not sure how to recycle single spike selection
        #    replot is calling self.clipPlot.clear()
        if self.singleSpikeMultiLine is not None:
            self.clipPlot.removeItem(self.singleSpikeMultiLine)
        if spikeNumber is not None:
            # only add i fwe have a spike
            self.singleSpikeMultiLine = MultiLine(
                x,
                y,
                self,
                width=3,
                allowXAxisDrag=False,
                forcePenColor="y",
                type="spike selection",
            )
            self.clipPlot.addItem(self.singleSpikeMultiLine)

    def selectSpikeList(self):
        """Select spikes based on self.getSelectedSpikes().

        sDict (dict): NOT USED
        """
        # logger.info(sDict)

        x = np.zeros(1) * np.nan
        y = np.zeros(1) * np.nan

        _selectedSpikes = self.getSelectedSpikes()

        if len(_selectedSpikes) > 0:
            try:
                x = self.x[_selectedSpikes]
                y = self.y[_selectedSpikes]
            except IndexError as e:
                pass
        if self.spikeListMultiLine is not None:
            self.clipPlot.removeItem(self.spikeListMultiLine)
        if len(_selectedSpikes) > 0:
            self.spikeListMultiLine = MultiLine(
                x,
                y,
                self,
                width=3,
                allowXAxisDrag=False,
                forcePenColor="c",
                type="spike list selection",
            )
            self.clipPlot.addItem(self.spikeListMultiLine)


# class MultiLine(pg.QtGui.QGraphicsPathItem):
class MultiLine(QtWidgets.QGraphicsPathItem):
    """
    This will display a time-series whole-cell recording efficiently
    It does this by converting the array of points to a QPath

    see: https://stackoverflow.com/questions/17103698/plotting-large-arrays-in-pyqtgraph/17108463#17108463
    """

    def __init__(
        self,
        x,
        y,
        detectionWidget,
        type,
        width=1,
        forcePenColor=None,
        allowXAxisDrag=True,
    ):
        """
        x and y are 2D arrays of shape (Nplots, Nsamples)
        type: (dvdt, vm)
        """

        self.exportWidgetList = []

        self.x = x
        self.y = y
        self.detectionWidget = detectionWidget
        self.myType = type
        self.allowXAxisDrag = allowXAxisDrag

        self.xDrag = None  # if true, user is dragging x-axis, otherwise y-axis
        self.xStart = None
        self.xCurrent = None
        self.linearRegionItem = None

        if len(x.shape) == 2:
            connect = np.ones(x.shape, dtype=bool)
            connect[:, -1] = 0  # don't draw the segment between each trace
            self.path = pg.arrayToQPath(x.flatten(), y.flatten(), connect.flatten())
        else:
            self.path = pg.arrayToQPath(x.flatten(), y.flatten(), connect="all")
        # pg.QtGui.QGraphicsPathItem.__init__(self, self.path)
        super().__init__(self.path)

        if forcePenColor is not None:
            penColor = forcePenColor
        else:
            penColor = "k"
        #
        if self.myType == "meanclip":
            penColor = "r"
            width = 3

        # print('MultiLine penColor:', penColor)
        # logger.info(f'{self.myType} penColor: {penColor}')
        pen = pg.mkPen(color=penColor, width=width)

        # testing gradient pen
        """
        cm = pg.colormap.get('CET-L17') # prepare a linear color map
        #cm.reverse()                    # reverse it to put light colors at the top
        pen = cm.getPen( span=(0.0,1.0), width=5 ) # gradient from blue (y=0) to white (y=1)
        """

        # this works
        self.setPen(pen)

    def shape(self):
        # override because QGraphicsPathItem.shape is too expensive.
        # print(time.time(), 'MultiLine.shape()', pg.QtGui.QGraphicsItem.shape(self))
        # return pg.QtGui.QGraphicsItem.shape(self)
        # now this
        return super().shape()

    def boundingRect(self):
        # print(time.time(), 'MultiLine.boundingRect()', self.path.boundingRect())
        return self.path.boundingRect()

    def old_mouseClickEvent(self, event):
        if event.button() == QtCore.Qt.RightButton:
            print("mouseClickEvent() right click", self.myType)
            self.contextMenuEvent(event)

    def old_contextMenuEvent(self, event):
        myType = self.myType

        if myType == "clip":
            print("WARNING: no export for clips, try clicking again")
            return

        """
        contextMenu = QtWidgets.QMenu()
        exportTraceAction = contextMenu.addAction(f'Export Trace {myType}')
        contextMenu.addSeparator()
        resetAllAxisAction = contextMenu.addAction(f'Reset All Axis')
        resetYAxisAction = contextMenu.addAction(f'Reset Y-Axis')
        #openAct = contextMenu.addAction("Open")
        #quitAct = contextMenu.addAction("Quit")
        #action = contextMenu.exec_(self.mapToGlobal(event.pos()))
        posQPoint = QtCore.QPoint(event.screenPos().x(), event.screenPos().y())
        action = contextMenu.exec_(posQPoint)
        if action is None:
            return
        actionText = action.text()
        if actionText == f'Export Trace {myType}':
            #print('Opening Export Trace Window')

            if self.myType == 'vmFiltered':
                xyUnits = ('Time (sec)', 'Vm (mV)')# todo: pass xMin,xMax to constructor
            elif self.myType == 'dvdtFiltered':
                xyUnits = ('Time (sec)', 'dV/dt (mV/ms)')# todo: pass xMin,xMax to constructor
            elif self.myType == 'meanclip':
                xyUnits = ('Time (ms)', 'Vm (mV)')# todo: pass xMin,xMax to constructor
            else:
                logger.error(f'Unknown myType: "{self.myType}"')
                xyUnits = ('error time', 'error y')

            path = self.detectionWidget.ba.path

            xMin = None
            xMax = None
            if self.myType in ['clip', 'meanclip']:
                xMin, xMax = self.detectionWidget.clipPlot.getAxis('bottom').range
            else:
                xMin, xMax = self.detectionWidget.getXRange()
            #print('  xMin:', xMin, 'xMax:', xMax)

            if self.myType in ['vm', 'dvdt']:
                xMargin = 2 # seconds
            else:
                xMargin = 2

            exportWidget = sanpy.interface.bExportWidget(self.x, self.y,
                            xyUnits=xyUnits,
                            path=path,
                            xMin=xMin, xMax=xMax,
                            xMargin = xMargin,
                            type=self.myType,
                            darkTheme=self.detectionWidget.useDarkStyle)

            exportWidget.myCloseSignal.connect(self.slot_closeChildWindow)
            exportWidget.show()

            self.exportWidgetList.append(exportWidget)
        elif actionText == 'Reset All Axis':
            #print('Reset Y-Axis', self.myType)
            self.detectionWidget.setAxisFull()
        elif actionText == 'Reset Y-Axis':
            #print('Reset Y-Axis', self.myType)
            self.detectionWidget.setAxisFull_y(self.myType)
        else:
            logger.warning(f'action not taken: {action}')
        """

    def old_slot_closeChildWindow(self, windowPointer):
        # print('closeChildWindow()', windowPointer)
        # print('  exportWidgetList:', self.exportWidgetList)

        idx = self.exportWidgetList.index(windowPointer)
        if idx is not None:
            popedItem = self.exportWidgetList.pop(idx)
            # print('  popedItem:', popedItem)
        else:
            print(" slot_closeChildWindow() did not find", windowPointer)

    def old_mouseDragEvent(self, ev):
        """
        default is to drag x-axis, use alt_drag for y-axis
        """
        # print('MultiLine.mouseDragEvent():', type(ev), ev)

        if ev.button() != QtCore.Qt.LeftButton:
            ev.ignore()
            return

        # modifiers = QtWidgets.QApplication.keyboardModifiers()
        # isAlt = modifiers == QtCore.Qt.AltModifier
        isAlt = ev.modifiers() == QtCore.Qt.AltModifier

        # xDrag = not isAlt # x drag is dafault, when alt is pressed, xDrag==False

        # allowXAxisDrag is now used for both x and y
        if not self.allowXAxisDrag:
            ev.accept()  # this prevents click+drag of plot
            return

        if ev.isStart():
            self.xDrag = not isAlt
            if self.xDrag:
                self.xStart = ev.buttonDownPos()[0]
                self.linearRegionItem = pg.LinearRegionItem(
                    values=(self.xStart, 0), orientation=pg.LinearRegionItem.Vertical
                )
            else:
                # in y-drag, we need to know (vm, dvdt)
                self.xStart = ev.buttonDownPos()[1]
                self.linearRegionItem = pg.LinearRegionItem(
                    values=(0, self.xStart), orientation=pg.LinearRegionItem.Horizontal
                )
            # self.linearRegionItem.sigRegionChangeFinished.connect(self.update_x_axis)
            # add the LinearRegionItem to the parent widget (Cannot add to self as it is an item)
            self.parentWidget().addItem(self.linearRegionItem)
        elif ev.isFinish():
            if self.xDrag:
                set_xyBoth = "xAxis"
            else:
                set_xyBoth = "yAxis"
            # self.parentWidget().setXRange(self.xStart, self.xCurrent)
            self.detectionWidget.setAxis(
                self.xStart, self.xCurrent, set_xyBoth=set_xyBoth, whichPlot=self.myType
            )

            self.xDrag = None
            self.xStart = None
            self.xCurrent = None

            """
            if self.myType == 'clip':
                if self.xDrag:
                    self.clipPlot.setXRange(self.xStart, self.xCurrent, padding=padding)
                else:
                    self.clipPlot.setYRange(self.xStart, self.xCurrent, padding=padding)
            else:
                self.parentWidget().removeItem(self.linearRegionItem)
            """
            self.parentWidget().removeItem(self.linearRegionItem)

            self.linearRegionItem = None

            return

        if self.xDrag:
            self.xCurrent = ev.pos()[0]
            # print('xStart:', self.xStart, 'self.xCurrent:', self.xCurrent)
            self.linearRegionItem.setRegion((self.xStart, self.xCurrent))
        else:
            self.xCurrent = ev.pos()[1]
            # print('xStart:', self.xStart, 'self.xCurrent:', self.xCurrent)
            self.linearRegionItem.setRegion((self.xStart, self.xCurrent))
        ev.accept()


if __name__ == "__main__":
    # path = '/home/cudmore/Sites/SanPy/data/19114001.abf'
    # path = '/home/cudmore/Sites/SanPy/data/19114001.abf'
    path = "/media/cudmore/data/rabbit-ca-transient/jan-12-2022/Control/220110n_0003.tif.frames/220110n_0003.tif"
    ba = sanpy.bAnalysis(path)

    """
    detectionClass = sanpy.bDetection()
    print(type(detectionClass))
    detectionClass['dvdtThreshold'] = math.nan #if None then detect only using mvThreshold
    detectionClass['mvThreshold'] = -0 #0.5
    """

    mvThreshold = 1.2  # 0
    detectionClass = sanpy.bDetection()  # gets default detection class
    detectionType = sanpy.bDetection.detectionTypes.mv
    detectionClass[
        "detectionType"
    ] = detectionType  # set detection type to ('dvdt', 'vm')
    detectionClass["dvdtThreshold"] = math.nan
    detectionClass["mvThreshold"] = mvThreshold
    detectionClass["preSpikeClipWidth_ms"] = 100
    detectionClass["postSpikeClipWidth_ms"] = 1000

    ba.spikeDetect(detectionClass=detectionClass)
    # ba.spikeDetect()
    print(ba)

    # sys.exit(1)

    import sys

    app = QtWidgets.QApplication([])
    sc = spikeClips(ba=ba)
    sc.show()
    sys.exit(app.exec_())



# 20210619

from PyQt5 import QtCore, QtWidgets, QtGui

from sanpy.sanpyLogger import get_logger

logger = get_logger(__name__)

import sanpy
from sanpy.interface.plugins import sanpyPlugin
# from sanpy.interface.plugins import ResponseType


class SummarizeResults(sanpyPlugin):
    """Plugin to display summary of analysis.

    Three versions are included
     - All: similar to what is exported to csv
     - Human Readable: A subset of human redable columns
     - Sweep summary: Report the mean/std/se/n etc for one sweep

    Uses:
        sanpy.bExport
        QTableView: sanpy.interface.bErrorTable.errorTableView()
        QAbstractTableModel: sanpy.interface.bFileTable.pandasModel
    """

    myHumanName = "Summarize Results"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # self.toggleResponseOptions(ResponseType.setSweep, False)  # we plot all sweeps
        # self.toggleResponseOptions(ResponseType.setAxis, False)

        self._reportType = 'Full Export'
        
        self.cardiacDf = None

        self._buildUI()

        self.replot()

    def _buildUI(self):

        self._buildingInterface = True

        layout = QtWidgets.QVBoxLayout()

        # one row of controls
        controlsLayout = QtWidgets.QHBoxLayout()
        layout.addLayout(controlsLayout)

        # the number of spikes in the report
        aLabel = QtWidgets.QLabel("Report Type")
        controlsLayout.addWidget(aLabel)

        #groupBox = QtWidgets.QGroupBox("Report Type")
        radio1 = QtWidgets.QRadioButton("Full Export")
        radio1.toggled.connect(lambda:self._on_radio_clicked(radio1))

        radio2 = QtWidgets.QRadioButton("Human Readable")
        radio2.toggled.connect(lambda:self._on_radio_clicked(radio2))

        radio3 = QtWidgets.QRadioButton("Sweep Summary")
        radio3.toggled.connect(lambda:self._on_radio_clicked(radio3))

        radio4 = QtWidgets.QRadioButton("Detection Errors")
        radio4.toggled.connect(lambda:self._on_radio_clicked(radio4))

        radio1.setChecked(True)

        controlsLayout.addWidget(radio1)
        controlsLayout.addWidget(radio2)
        controlsLayout.addWidget(radio3)
        controlsLayout.addWidget(radio4)

        # the number of spikes in the report
        self.numSpikesLabel = QtWidgets.QLabel("no spikes")
        controlsLayout.addWidget(self.numSpikesLabel)

        self.myErrorTable = sanpy.interface.bErrorTable.errorTableView()
        # TODO: derive a more general purpose table, here we are re-using error table
        self.myErrorTable.horizontalHeader().setSectionResizeMode(
            QtWidgets.QHeaderView.Interactive
        )
        self.myErrorTable.horizontalHeader().setStretchLastSection(False)

        layout.addWidget(self.myErrorTable)

        # self.setLayout(layout)
        self.getVBoxLayout().addLayout(layout)

        #
        # connect clicks in error table to signal main sanpy_app with slot_selectSpike()
        logger.warning("mar 11 FIX SPIKE SELECTION")
        if self.getSanPyApp() is not None:
            fnPtr = self.getSanPyApp().slot_selectSpike
            self.myErrorTable.signalSelectSpike.connect(fnPtr)

        self._buildingInterface = False

    def _on_radio_clicked(self, radioButton):
        """Respond to user click in radio buttons.
        
        We receive this twice on each new radio button selection
        We receive false for what gets unselected and then true for the new selection.
        
        Only prococess True
        """
        if self._buildingInterface:
            return
        
        buttonName = radioButton.text()
        isChecked = radioButton.isChecked()
        logger.info(f'buttonName:{buttonName} isChecked:{isChecked}')
        if isChecked:
            self._reportType = buttonName
            self.replot()

    def setAxis(self):
        # inherited, resopnd to user setting x-axis
        self.replot()

    def replot(self):
        """Generate report with sanpy.bExport.report2()
        and put it into a table.
        """
        logger.info("")
        # update
        if self.ba is None:
            return

        # an export object to create any number of reports
        exportObject = sanpy.bExport(self.ba)

        startSec, stopSec = self.getStartStop()

        self.cardiacDf = None

        if self._reportType == 'Full Export':
            # v2, full export
            self.cardiacDf = exportObject.report3(
                sweep = self.sweepNumber,
                epoch = self.epochNumber,
                theMin=startSec,
                theMax=stopSec
            )  # report2 is more 'cardiac'
        
        elif self._reportType == 'Human Readable':
            # v1, human readable columns
            self.cardiacDf = exportObject.report2(
                sweep = self.sweepNumber,
                epoch = self.epochNumber,
                theMin=startSec,
                theMax=stopSec
            )  # report2 is more 'cardiac'

        elif self._reportType == 'Sweep Summary':
            # v3
            self.cardiacDf = exportObject.getSummary(
                sweep=self.sweepNumber,
                epoch=self.epochNumber,
                theMin=startSec,
                theMax=stopSec)

        elif self._reportType == 'Detection Errors':
            self.cardiacDf = self.ba.dfError

        else:
            logger.warning(f'Did not understand _reportType:{self._reportType}')

        if self.cardiacDf is None:
            return
    
        errorReportModel = sanpy.interface.bFileTable.pandasModel(self.cardiacDf)
        self.myErrorTable.setModel(errorReportModel)

        for colIdx in range(len(self.cardiacDf.columns)):
            if colIdx == 0:
                continue
            self.myErrorTable.setColumnWidth(colIdx, 170)

        self.numSpikesLabel.setText(f"{len(self.cardiacDf)} spikes")

    def copyToClipboard(self):
        if self.ba is not None and self.cardiacDf is not None:
            logger.info("Copy to clipboard")
            self.cardiacDf.to_clipboard(sep="\t", index=False)

    def selectSpikeList(self):
        spikeList = self.getSelectedSpikes()
        logger.info(f"{spikeList}")
        self.myErrorTable.mySelectRows(spikeList)


def main():
    path = "/home/cudmore/Sites/SanPy/data/19114001.abf"
    ba = sanpy.bAnalysis(path)
    ba.spikeDetect()
    print(ba.numSpikes)

    import sys

    app = QtWidgets.QApplication([])
    rt = SummarizeResults(ba=ba)
    rt.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



# 20210522

# from PyQt5 import QtCore, QtWidgets, QtGui
from PyQt5 import QtWidgets

"""
General purpose dialogs
"""


def okDialog(message, informativeText=None):
    """
    message:
    """
    msg = QtWidgets.QMessageBox()
    msg.setIcon(QtWidgets.QMessageBox.Warning)
    msg.setText(message)
    if informativeText is not None:
        msg.setInformativeText(informativeText)
    msg.setWindowTitle(message)
    retval = msg.exec_()
    return retval


def okCancelDialog(message, informativeText=None):
    msg = QtWidgets.QMessageBox()
    msg.setIcon(QtWidgets.QMessageBox.Warning)
    msg.setText(message)
    if informativeText is not None:
        msg.setInformativeText(informativeText)
    msg.setWindowTitle(message)
    msg.setStandardButtons(QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel)
    msg.setDefaultButton(QtWidgets.QMessageBox.Ok)
    retval = msg.exec_()
    return retval


def yesNoCancelDialog(message, informativeText=None):
    msg = QtWidgets.QMessageBox()
    msg.setIcon(QtWidgets.QMessageBox.Warning)
    msg.setText(message)
    if informativeText is not None:
        msg.setInformativeText(informativeText)
    msg.setWindowTitle(message)
    msg.setStandardButtons(
        QtWidgets.QMessageBox.Yes
        | QtWidgets.QMessageBox.No
        | QtWidgets.QMessageBox.Cancel
    )
    retval = msg.exec_()
    return retval



import numpy as np
import scipy.signal

from PyQt5 import QtCore, QtGui, QtWidgets

import matplotlib.pyplot as plt

from sanpy.sanpyLogger import get_logger

logger = get_logger(__name__)

import sanpy
from sanpy.interface.plugins import sanpyPlugin

from sanpy.bAnalysisUtil import statList


class basePlotTool(sanpyPlugin):
    """ """

    myHumanName = "Base Plot Tool"

    def __init__(self, **kwargs):
        super(basePlotTool, self).__init__(**kwargs)

        self.masterDf = None
        # one
        # self.masterDf = self.ba.dfReportForScatter
        # pool of bAnalysis
        # self.masterDf = self._bPlugins._sanpyApp.myAnalysisDir.pool_build()

        self.mainWidget2 = None

        # turn off all signal/slot
        switchFile = self.responseTypes.switchFile
        self.toggleResponseOptions(switchFile, newValue=False)
        analysisChange = self.responseTypes.analysisChange
        self.toggleResponseOptions(analysisChange, newValue=False)
        selectSpike = self.responseTypes.selectSpike
        self.toggleResponseOptions(selectSpike, newValue=False)
        setAxis = self.responseTypes.setAxis
        self.toggleResponseOptions(setAxis, newValue=False)

        # self.plot()

    def plot(self):
        logger.info("")

        # this plugin does not rely on underlying self.ba
        # if self.ba is None:
        #    return

        # analysisName = 'analysisname'
        analysisName = "file"
        statListDict = statList  # maps human readable to comments
        categoricalList = [
            "include",
            "condition",
            "cellType",
            "sex",
            "file",
            "File Number",
        ]  # , 'File Name']
        hueTypes = [
            "cellType",
            "sex",
            "condition",
            "file",
            "File Number",
        ]  # , 'File Name'] #, 'None']
        sortOrder = ["cellType", "sex", "condition", "File Number"]

        limitToCol = ["epoch"]

        interfaceDefaults = {
            "Y Statistic": "Spike Frequency (Hz)",
            "X Statistic": "Spike Frequency (Hz)",
            "Hue": "cellType",
            "Group By": "file",
        }

        # analysisName, masterDf = analysisName, df0 = ba.getReportDf(theMin, theMax, savefile)

        # print('!!!! FIX THIS in plugin plotTool.plot()')
        # masterDf = self._bPlugins._sanpyApp.myAnalysisDir.pool_build()
        # was this
        # masterDf = self.ba.dfReportForScatter

        if self.masterDf is None:
            logger.error("Did not get analysis df, be sure to run detectioon")
            return
        # bScatterPlotMainWindow
        # self.scatterWindow = sanpy.scatterwidget.bScatterPlotMainWindow(
        path = ""
        self.mainWidget2 = sanpy.interface.bScatterPlotMainWindow(
            path,
            categoricalList,
            hueTypes,
            analysisName,
            sortOrder,
            statListDict=statListDict,
            masterDf=self.masterDf,
            limitToCol=limitToCol,
            interfaceDefaults=interfaceDefaults,
        )
        # rewire existing widget into plugin architecture
        # self.mainWidget.closeEvent = self.onClose
        self._mySetWindowTitle()

        # self.mainWidget2.show()

        tmpLayout = QtWidgets.QVBoxLayout()
        tmpLayout.addWidget(self.mainWidget2)

        #
        # self.setLayout(tmpLayout)
        self.getVBoxLayout().addLayout(tmpLayout)



import time
from typing import Union, Dict, List, Tuple, Optional, Optional
from functools import partial

import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
import seaborn as sns

from PyQt5 import QtCore, QtWidgets, QtGui

from sanpy.sanpyLogger import get_logger

logger = get_logger(__name__)

import sanpy
from sanpy import bDetection
from sanpy import bAnalysis
from sanpy.interface import myTableView_tmp  # name conflict with interface.myTableView
from sanpy.interface.plugins import sanpyPlugin
from sanpy.interface.plugins import (
    ResponseType,
)  # to toggle response to set sweeps, etc
from sanpy.interface.plugins import myStatListWidget


def getStatFi(
    dfMaster: pd.DataFrame,
    stat: str,
    epochNumber: int = 2,
    intervalStat: bool = False,
    filename: str = "",
) -> pd.DataFrame:
    """Given a bAnalysis stat, derive a number of stats
    for each sweep

    Sweeps correspond to different current injection amplitudes.

    Parameters
    ----------
    dfMaster : DataFrame
    stat : str
    epochNumber : int
    epochNumber : bool
        If true then skip first spike.
        Used for stat like spike frequency (Hz) or ISI (ms)
    filename : str
        Append filename column

    Notes
    -----
    Rows in the table are sweeps, if there are no spikes in a sweep it will not be included
    """

    _startSec = time.time()

    # reduce to only spikes for given epochNumber
    dfEpoch = dfMaster[dfMaster["epoch"] == epochNumber]

    # epochLevels is a list of current injection amplitudes
    epochLevels = dfEpoch["epochLevel"].unique()
    epochLevels = np.round(epochLevels, 3)

    # we also want 1st, 2nd, 3rd and nth values
    # don't use agg variable `first` or `last`` (they skip nan)
    #   `second`` is not available
    agg_func_math = {stat: ["count", "mean", "median", "min", "max", "std", "sem"]}
    as_index = True  # if True then 'sweep' number becomes the row index
    dfSweepsSummary = (
        dfEpoch.groupby(["sweep"], as_index=as_index).agg(agg_func_math).round(2)
    )

    # flatten
    dfSweepsSummary.columns = [
        "_".join(col).rstrip("_") for col in dfSweepsSummary.columns.values
    ]

    # insert as first column, after 'sweeps'
    dfSweepsSummary.insert(0, "epochLevel", epochLevels)
    # dfSweepsSummary['epochLevel'] = epochLevels

    # get the first, second and last values
    # if we are doing interval statistics like 'inter spike interval'
    # then always ignore the `first` event in the epoch (e.g. no interval for first spike)

    if intervalStat:
        firstSpike = 1
    else:
        firstSpike = 0

    _first = dfEpoch.groupby(["sweep"], as_index=True)[stat].nth(firstSpike)
    dfSweepsSummary.at[_first.index, stat + "_first"] = _first.values
    # print('_first.index', _first.index)

    _second = dfEpoch.groupby(["sweep"], as_index=True)[stat].nth(firstSpike + 1)
    dfSweepsSummary.at[_second.index, stat + "_second"] = _second.values
    # print('_second.index', _second.index)

    # _third = dfEpoch.groupby(['sweep'], as_index=True)[stat].nth(2)
    # dfSweepsSummary.at[_second.index, stat + '_third'] = _third.values

    _last = dfEpoch.groupby(["sweep"], as_index=True)[stat].nth(-1)
    dfSweepsSummary.at[_last.index, stat + "_last"] = _last.values
    # print('_last.index', _last.index)

    # 1st / 2nd and 1s / last
    dfSweepsSummary[stat + "_1_2"] = (
        dfSweepsSummary[stat + "_first"] / dfSweepsSummary[stat + "_second"]
    )
    dfSweepsSummary[stat + "_1_n"] = (
        dfSweepsSummary[stat + "_first"] / dfSweepsSummary[stat + "_last"]
    )

    # coefficient of variation.
    # Always calculate but only meaningful for interval stats?
    dfSweepsSummary[stat + "_cv"] = (
        dfSweepsSummary[stat + "_std"] / dfSweepsSummary[stat + "_mean"]
    )

    dfSweepsSummary["filename"] = filename

    # move index column 'sweep' into a column
    dfSweepsSummary = dfSweepsSummary.reset_index()

    _stopSec = time.time()
    # print(f'Took {round(_stopSec-_startSec,3)} seconds')

    return dfSweepsSummary


class plotFi(sanpyPlugin):
    myHumanName = "Plot FI"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.toggleResponseOptions(ResponseType.setSweep, False)  # we plot all sweeps
        self.toggleResponseOptions(ResponseType.setAxis, False)

        # inherited property
        self._epochNumber = 2

        self.setDefaultPlot()

        self.df_fi = None
        self._buildUI()

        self._refreshPlotOptionsLayout()

        self._selectInTable("Spike Frequency (Hz)")

        self.replot()

    def setDefaultPlot(self):
        """Default plot dict"""

        self._plotDict = {
            "Raw": True,
            "Error": "sem",  # in [None, 'std', 'sem']
            "Legend": False,
            # keys correspond to columns returned by getStatFI()
            "overlays": {
                "Mean": True,
                "First": False,
                "Last": False,
                "CV": False,
                "Count": False,
                "1_2": False,  # first spike / second spike
                "1_n": False,  # first spike / last spike
            },
        }

    def on_error_combo_box(self, item: str):
        if item == "None":
            item = None
        self._plotDict["Error"] = item
        self.replot()

    def _selectInTable(self, findStatKey: str):
        # this is a CLUDGE !!! FIX IT
        # select in table
        _statList = self._yStatListWidget.statList
        # find 'Spike Frequency (Hz)'
        foundKeyIdx = None
        for keyIdx, k in enumerate(_statList.keys()):
            if k == findStatKey:
                foundKeyIdx = keyIdx
        if foundKeyIdx is not None:
            self._yStatListWidget.myTableWidget.selectRow(foundKeyIdx)

    def on_button_click(self, name: str):
        """
        Parameters
        ----------
        name : str
            Name of the buton clicked
        """
        logger.info(name)
        if name == "Inst Freq":
            self.setDefaultPlot()  # reset _plotDict to defaults
            # set based on button
            self._plotDict["Raw"] = True
            self._plotDict["Error"] = "sem"
            self._plotDict["overlays"]["Mean"] = True

            self._selectInTable("Spike Frequency (Hz)")

        elif name == "Spike Count":
            self.setDefaultPlot()  # reset _plotDict to defaults
            # set based on button
            self._plotDict["Raw"] = False
            self._plotDict["Error"] = None
            self._plotDict["Legend"] = True
            self._plotDict["overlays"]["Count"] = True
        else:
            logger.warning(f'Did not understand "{name}')
            return

        # refresh interface
        self._refreshPlotOptionsLayout()

        # replot
        self.replot()

    def on_check_click(self, state: int, name: str):
        isChecked = state > 0

        logger.info(f"{name} is checked {isChecked}")

        _doReplot = False

        if name == "Results Table":
            self._fiTableView.setVisible(isChecked)
        elif name == "Plot Toolbar":
            self._mplToolbar.setVisible(isChecked)
        elif name in self._plotDict.keys():
            # top level like [Raw, Legend, Error]
            self._plotDict[name] = isChecked
            _doReplot = True
        elif name in self._plotDict["overlays"].keys():
            self._plotDict["overlays"][name] = isChecked
            _doReplot = True
        else:
            logger.warning(f'Did not understand checkbox "{name}"')
            return

        if _doReplot:
            self.replot()

    def _buildTopToolbar(self):
        _topToolbar = QtWidgets.QHBoxLayout()

        _defualtButtons = ["Inst Freq", "Spike Count"]

        for buttonName in _defualtButtons:
            _aButton = QtWidgets.QPushButton(buttonName)
            _aButton.clicked.connect(partial(self.on_button_click, buttonName))
            _topToolbar.addWidget(_aButton, alignment=QtCore.Qt.AlignLeft)

        # toggle the fi stat table
        name = "Results Table"
        _aCheckbox = QtWidgets.QCheckBox(name)
        _aCheckbox.setChecked(False)
        _aCheckbox.stateChanged.connect(
            lambda state, name=name: self.on_check_click(state, name)
        )
        _topToolbar.addWidget(_aCheckbox, alignment=QtCore.Qt.AlignLeft)

        # toggle self._mplToolbar
        name = "Plot Toolbar"
        _aCheckbox = QtWidgets.QCheckBox(name)
        _aCheckbox.setChecked(False)
        _aCheckbox.stateChanged.connect(
            lambda state, name=name: self.on_check_click(state, name)
        )
        _topToolbar.addWidget(_aCheckbox, alignment=QtCore.Qt.AlignLeft)

        _topToolbar.addStretch()

        return _topToolbar

    def _refreshPlotOptionsLayout(self):
        """Refresh all buttons in _overlayCheckboxes with self._plotDict['overlays']"""

        # raw
        _plotRaw = self._plotDict["Raw"]
        self._rawCheckbox.setChecked(_plotRaw)

        # error
        _errorType = self._plotDict["Error"]
        if _errorType is None:
            _errorType = "None"
        self._errorComboBox.setCurrentText(_errorType)

        for name, value in self._plotDict["overlays"].items():
            self._overlayCheckBoxes[name].setChecked(value)

    def _buildPlotOptionsLayout(self):
        _vLayout = QtWidgets.QVBoxLayout()

        # to toggle columns in fi analysis df

        name = "Legend"
        value = self._plotDict[name]
        _aCheckbox = QtWidgets.QCheckBox(name)
        _aCheckbox.setChecked(value)
        _aCheckbox.stateChanged.connect(
            lambda state, name=name: self.on_check_click(state, name)
        )
        _vLayout.addWidget(_aCheckbox, alignment=QtCore.Qt.AlignTop)

        name = "Raw"
        value = self._plotDict[name]
        self._rawCheckbox = QtWidgets.QCheckBox(name)
        self._rawCheckbox.setChecked(value)
        self._rawCheckbox.stateChanged.connect(
            lambda state, name=name: self.on_check_click(state, name)
        )
        _vLayout.addWidget(self._rawCheckbox, alignment=QtCore.Qt.AlignTop)

        name = "Error"
        _errorLayout = QtWidgets.QHBoxLayout()
        _aLabel = QtWidgets.QLabel(name)
        _errorLayout.addWidget(_aLabel)
        self._errorComboBox = QtWidgets.QComboBox()
        self._errorComboBox.addItems(["None", "std", "sem"])
        self._errorComboBox.currentTextChanged.connect(self.on_error_combo_box)
        _errorLayout.addWidget(self._errorComboBox, alignment=QtCore.Qt.AlignTop)
        _vLayout.addLayout(_errorLayout)

        self._overlayCheckBoxes = {}
        for name, value in self._plotDict["overlays"].items():
            _aCheckbox = QtWidgets.QCheckBox(name)
            _aCheckbox.setChecked(value)
            _aCheckbox.stateChanged.connect(
                lambda state, name=name: self.on_check_click(state, name)
            )
            _vLayout.addWidget(_aCheckbox, alignment=QtCore.Qt.AlignTop)

            self._overlayCheckBoxes[name] = _aCheckbox

        _vLayout.addStretch()

        return _vLayout

    def replot(self):
        """Replot when file or analysis changes."""

        logger.info("")

        if self.ba is None:
            return

        # get from stat lists
        yHumanStat, yStat = self._yStatListWidget.getCurrentStat()

        # always clear the axis
        self.axs.clear()

        self.axs.autoscale(enable=True, axis="y", tight=None)

        # get masterDf
        dfMaster = self.ba.spikeDict.asDataFrame()
        if dfMaster is None or dfMaster.empty:
            # no spikes
            logger.info("no spikes to plot -> return")
            return

        # user has to specify which epoch
        plotRaw = self._plotDict["Raw"]
        if plotRaw:
            # reduce to only spikes for given epochNumber
            dfEpoch = dfMaster[dfMaster["epoch"] == self.epochNumber]
            # sns.scatterplot(x='epochLevel', y=stat, data=dfEpoch, ax=_ax)
            self.axs.scatter(
                "epochLevel",
                yStat,
                data=dfEpoch,
                marker="o",
                c="#aaaaaa",
                label=yHumanStat,
            )

        #
        # get a df with summary fi data
        intervalStat = False
        if yStat in ["spikeFreq_hz", "isi_ms"]:
            intervalStat = True
        _filename = self.ba.getFileName()
        logger.info(
            f"calling getStatFi() with epoch:{self.epochNumber} intervalStat:{intervalStat} filename:{_filename}"
        )
        self.df_fi = getStatFi(
            dfMaster,
            yStat,
            epochNumber=self.epochNumber,
            intervalStat=intervalStat,
            filename=_filename,
        )

        logger.info("generated new fi table self.df_fi:")
        print(self.df_fi)

        if self.df_fi.empty:
            logger.warning("  got empty dataframe -> return")
            return

        self._fiTableView.slotSwitchTableDf(self.df_fi)

        # cycle color
        _fiStatColor = ["c", "m", "r", "b"] * 5

        # for plotIdx, _fiStat in enumerate(fiStatList):
        plotIdx = -1
        for _fiStat, value in self._plotDict["overlays"].items():
            plotIdx += 1
            if not value:
                # plot is turned off
                continue
            _stat = yStat + "_" + _fiStat.lower()
            # _ax = sns.scatterplot(x='epochLevel', y=_stat, data=df_fi, ax=_ax)
            # _ax.plot('epochLevel', _stat, data=df_fi, linestyle='-', marker='o', color='k')
            currentColor = _fiStatColor[plotIdx]
            logger.info(f'plotting _stat:"{_stat}"')
            _errorType = self._plotDict["Error"]
            if _fiStat == "Mean" and _errorType is not None:
                _yerr = yStat + "_" + _errorType
            else:
                _yerr = None
            self.axs.errorbar(
                "epochLevel",
                _stat,
                yerr=_yerr,
                data=self.df_fi,
                linestyle="-",
                marker="o",
                color=currentColor,
                label=_fiStat,
            )

        if self._plotDict["Legend"]:
            self.axs.legend(loc="upper left")

        #
        # set x-axis to full range of all injected currents (e.g. epochLevel)
        # pad by 10%
        epochLevels = dfMaster["epochLevel"].unique()
        _minEpochLevel = np.min(epochLevels)
        _maxEpochLevel = np.max(epochLevels)
        _xRange = _maxEpochLevel - _minEpochLevel
        _xPad = _xRange * 0.1
        _xMin = _minEpochLevel - _xPad
        _xMax = _maxEpochLevel + _xPad
        logger.info(f"_xMin:{_xMin} _xMax:{_xMax}")
        if np.isnan(_xMin) or np.isnan(_xMax):
            pass
        else:
            self.axs.set_xlim([_xMin, _xMax])

        # set axis labels
        self.axs.set_xlabel("Current Step (pA)")
        self.axs.set_ylabel(yHumanStat)

        self.axs.spines["right"].set_visible(False)
        self.axs.spines["top"].set_visible(False)

        # redraw
        self.static_canvas.draw()

    def copyToClipboard(self, df: pd.DataFrame = None):
        logger.info("")
        if self.df_fi is None:
            return
        self.df_fi.to_clipboard(sep="\t", index=False)

        # want to provide some feedback
        # #if self.getSanPyApp() is not None:

    def _buildUI(self):
        # top toolbar
        _topToolbar = self._buildTopToolbar()
        self.getVBoxLayout().addLayout(_topToolbar)

        main_hLayout = QtWidgets.QHBoxLayout()

        # add list of y stats
        # has statList, a copy of sanpy.bAnalysisUtil.getStatList()
        # looks like
        #   statList['Spike Time (s)'] = {
        #       'name': 'thresholdSec',

        # trim down stat list
        _statList = sanpy.bAnalysisUtil.getStatList()
        # for k in _statList.keys():
        #     print(k)
        _hideTheseKeys = [
            "Spike Number",
            "Sweep Spike Number",
            "Sweep Number",
            "Epoch",
            "Epoch DAC",
            "Epoch Spike Number",
        ]
        _statListShort = {}
        for k, v in _statList.items():
            if k not in _hideTheseKeys:
                _statListShort[k] = v
        # _statListShort = None
        self._yStatListWidget = myStatListWidget(
            self, statList=_statListShort, headerStr="Y-Stat"
        )
        self._yStatListWidget.myTableWidget.selectRow(3)
        main_hLayout.addWidget(self._yStatListWidget)

        # add vertical list of button for fi stats like (first, last, mean, cv, etc)
        _plotOptionsLayout = self._buildPlotOptionsLayout()
        main_hLayout.addLayout(_plotOptionsLayout)

        _vPlotLayout = QtWidgets.QVBoxLayout()
        _canvas, self._mplToolbar = self.mplWindow2(addToLayout=False)
        self._mplToolbar.setVisible(False)
        # self.axs.scatter('epochLevel', stat, data=dfEpoch, marker='o', c='#777777')
        self.axs.scatter([], [])
        sns.despine()

        _vPlotLayout.addWidget(_canvas)
        _vPlotLayout.addWidget(self._mplToolbar)

        main_hLayout.addLayout(_vPlotLayout)

        self.getVBoxLayout().addLayout(main_hLayout)

        #
        # table view of current fi stats
        self._fiTableView = myTableView_tmp("FI Results")
        self._fiTableView.setVisible(False)
        # self._fiTableView.slotSwitchTableDf(self.masterDf)
        self.getVBoxLayout().addWidget(self._fiTableView)


def testPlugin():
    # load raw data
    path = "data/2021_07_20_0010.abf"
    ba = bAnalysis(path)

    # spike detect
    _dDict = bDetection().getDetectionDict("Fast Neuron")
    ba.spikeDetect(_dDict)

    # run the plugin
    import sys

    app = QtWidgets.QApplication([])
    pfi = plotFi(ba=ba)
    pfi.show()
    sys.exit(app.exec_())


def test():
    # load raw data
    path = "data/2021_07_20_0010.abf"
    ba = bAnalysis(path)

    # spike detect
    _dDict = bDetection().getDetectionDict("Fast Neuron")
    ba.spikeDetect(_dDict)

    # get masterDf
    dfMaster = ba.asDataFrame()
    if dfMaster is None:
        # no spikes
        return

    # user has to specify which epoch
    epochNumber = 2

    # the plot we will append to
    fig, _ax = plt.subplots(1, 1, sharex=True)

    # decide on what we want to plot
    # this is done in gui

    #
    # fiStat could be [first, second, mean, count, etc)

    # fiStat is a column in df_fi (See below)

    # 1) plot num spikes, no raw data
    if 0:
        fiStat = "count"  # the number of spikes
        stat = "thresholdSec"  # thresholdSec, spikeFreq_hz
        plotRaw = False

    # 1.1) plot fiStat: first, last, mean, ... spike time
    if 0:
        fiStat = "first"
        fiStat = "last"
        fiStat = "mean"
        stat = "thresholdSec"  # thresholdSec, spikeFreq_hz
        plotRaw = True

    if 0:
        # 2) plot inst freq
        fiStat = (
            "second"  # spike freq is an interval statistic (first spike does not count)
        )
        fiStat = "mean"
        stat = "spikeFreq_hz"  # thresholdSec, spikeFreq_hz
        plotRaw = True

    if 0:
        # 3) plot isi_ms
        fiStat = (
            "second"  # spike freq is an interval statistic (first spike does not count)
        )
        stat = "isi_ms"  # thresholdSec, spikeFreq_hz
        plotRaw = True

    fiStatList = ["first", "last", "mean"]
    fiStatList = ["mean"]
    stat = "spikeFreq_hz"  # thresholdSec, spikeFreq_hz
    intervalStat = True
    plotRaw = True

    #
    # get a df with summary fi data
    df_fi = getStatFi(
        dfMaster, stat, epochNumber=epochNumber, intervalStat=intervalStat
    )

    print("=== summary df_fi:")
    print(df_fi)

    # plot raw data
    # lines is of type 'matplotlib.axes._subplots.AxesSubplot'
    if plotRaw:
        # reduce to only spikes for given epochNumber
        dfEpoch = dfMaster[dfMaster["epoch"] == epochNumber]
        # sns.scatterplot(x='epochLevel', y=stat, data=dfEpoch, ax=_ax)
        _ax.scatter("epochLevel", stat, data=dfEpoch, marker="o", c="#777777")

    #
    # overlay with one stat like _mean
    # order matters! This is on top of previous plots
    _yerr = stat + "_std"
    # _yerr = stat + '_sem'

    for _fiStat in fiStatList:
        _stat = stat + "_" + _fiStat
        # _ax = sns.scatterplot(x='epochLevel', y=_stat, data=df_fi, ax=_ax)
        # _ax.plot('epochLevel', _stat, data=df_fi, linestyle='-', marker='o', color='k')
        _ax.errorbar(
            "epochLevel",
            _stat,
            yerr=_yerr,
            data=df_fi,
            linestyle="-",
            marker="o",
            color="k",
        )

    #
    # set x-axis to full range of all injected currents (e.g. epochLevel)
    # TODO: expand a bit so we can see everything
    epochLevels = dfMaster["epochLevel"].unique()
    # epochLevels = np.round(epochLevels,3)
    _minEpochLevel = np.min(epochLevels)
    _maxEpochLevel = np.max(epochLevels)
    # pad by 10%
    _xRange = _maxEpochLevel - _minEpochLevel
    _xPad = _xRange * 0.1
    # set x-axis
    _ax.set_xlim([_minEpochLevel - _xPad, _maxEpochLevel + _xPad])

    sns.despine()

    plt.show()


if __name__ == "__main__":
    # test()
    testPlugin()



import os

import pandas as pd
import numpy as np

import sanpy

import sanpy.interface.plugins.stimGen2
from sanpy.interface.plugins.stimGen2 import readFileParams, buildStimDict

from sanpy.sanpyLogger import get_logger

logger = get_logger(__name__)

"""
bAnalysis class that loads a stim file.

TODO:
	- as usual, split into interface and gui.
		For example, move this to main api (move out of interface)
		sanpy.interface.plugins.stimGen2.readCommentParams(stimFileComment)
"""


class bAnalysisStim(sanpy.bAnalysis):
    def __init__(self, path, loadData=True, stimulusFileFolder=None):
        if stimulusFileFolder is None:
            stimulusFileFolder = os.path.split(path)[
                0
            ]  #'/media/cudmore/data/stoch-res' #os.path.split(path)[0]
        super(bAnalysisStim, self).__init__(
            file=path, loadData=True, stimulusFileFolder=stimulusFileFolder
        )

        print(self)
        print(self.abf)  # will be none
        # tmpSweepC = self.abf.sweepC

        # stimFilePath, stimFileComment = pyabf.stimulus.abbCommentFromFile(self.abf)
        stimFilePath, stimFileComment = self._commentFromAtfFile()
        self._stimDict = None  # raw parameters from snapy.interface.plugins.stimGen
        self._stimFile = None
        self._stimDictList = None  # one element per sweep
        self._stimFileDf = None
        if stimFileComment is not None:
            self._stimFile = stimFilePath
            self._stimDict = sanpy.interface.plugins.stimGen2.readCommentParams(
                stimFileComment
            )
            self._stimDict["file"] = os.path.split(stimFilePath)[
                1
            ]  # atf does not know abf file name
            # TODO: there is an error in saving atf comment string, post sweeps are always 0
            # print(f'self._stimDict:', self._stimDict)
            self._stimDictList = sanpy.interface.plugins.stimGen2.buildStimDict(
                self._stimDict, path=self.path
            )
            self._stimFileDf = self._stimFileAsDataFrame()

    def isiStats(self, hue="sweep"):
        """
        Get mean/sd/sem etc after grouping by 'sweep'
        """
        if self.numSpikes == 0:
            return None

        # df = self.analysisDf
        df = self.asDataFrame()
        # print('df:', df)
        if df is None:
            return pd.DataFrame()  # empty

        df = self.reduce()  # reduce based on start/stop of sin stimulus

        statList = []

        statStr = "isi_ms"

        stimFileDf = self._stimFileDf

        hueList = df[hue].unique()
        # for idx,oneHue in enumerate(hueList):
        for idx, sweep in enumerate(range(self.numSweeps)):
            oneHue = sweep

            dfPlot = df[df[hue] == oneHue]  # will be empty if 0 spikes in sweep

            # someSpikes = not dfPlot.empty # no spikes for this sweep

            ipi_ms = dfPlot[statStr]

            numIntervals = len(ipi_ms)
            # print('numIntervals:', numIntervals)

            minISI = np.min(ipi_ms) if numIntervals > 1 else float("nan")
            maxISI = np.max(ipi_ms) if numIntervals > 1 else float("nan")

            meanISI = np.nanmean(ipi_ms) if numIntervals > 1 else float("nan")
            medianISI = np.nanmedian(ipi_ms) if numIntervals > 1 else float("nan")
            stdISI = np.nanstd(ipi_ms) if numIntervals > 1 else float("nan")
            cvISI = stdISI / meanISI if numIntervals > 1 else float("nan")
            cvISI = round(cvISI, 3) if numIntervals > 1 else float("nan")
            cvISI_inv = (
                np.nanstd(1 / ipi_ms) / np.nanmean(1 / ipi_ms)
                if numIntervals > 1
                else float("nan")
            )
            cvISI_inv = round(cvISI_inv, 3) if numIntervals > 1 else float("nan")

            oneDict = {
                "file": self.getFileName(),
                "stat": statStr,
                "sweep": idx,
                "count": numIntervals,
                "minISI": minISI,  # round(np.nanmin(ipi_ms),3),
                "maxISI": maxISI,  # round(np.nanmax(ipi_ms),3),
                "stdISI": round(stdISI, 3),
                "meanISI": round(meanISI, 3),
                "medianISI": round(medianISI, 3),
                "cvISI": cvISI,
                "cvISI_inv": cvISI_inv,
                # stim params from stimulus file
                "Stim Freq (Hz)": "",
                "Stim Amp": "",
                "Noise Amp": "",
            }

            if stimFileDf is not None:
                if idx < len(stimFileDf):
                    oneDict["Stim Freq (Hz)"] = stimFileDf.loc[idx, "freq(Hz)"]
                    oneDict["Stim Amp"] = stimFileDf.loc[idx, "amp"]
                    oneDict["Noise Amp"] = stimFileDf.loc[idx, "noise amp"]
                else:
                    # This happens when pClamp sweeps > stimGen sweeps (like forgetting to set post=1 in stimgen)
                    logger.error(f"Did not find row {idx} in stimFileDict")

            statList.append(oneDict)

        #
        retDf = pd.DataFrame(statList)
        return retDf

    def _commentFromAtfFile(self, channel=0):
        """
        Get the comment string from atf stimulus file
        """
        from pyabf.stimulus import findStimulusWaveformFile
        from pyabf.stimulus import cachedStimuli

        stimPath = findStimulusWaveformFile(self.abf, channel)  # , verbose=False)
        if stimPath is None:
            return None, None
        else:
            return stimPath, cachedStimuli[stimPath].header["comment"]

    @property
    def stimDict(self):
        """
        If recorded with a stimulus file, return dict of stimulus parameters.
        Stimulus parameters are defined in sanpy/interface/plugins/stimGen2.py
        """
        return self._stimDict

    def reduce(self):
        """
        Reduce spikes based on start/stop of sin stimulus

        TODO:
                Use readFileParams(atfPath) to read stimulus ATF file and get start/stop

        Return:
                df of spikes within sin stimulus
        """
        # print(self.stimDict)

        # df = self.analysisDf
        df = self.asDataFrame()  # if no spikes, will be empty

        startSec, stopSec = self.getStimStartStop()

        if startSec is not None and stopSec is not None:
            df = df[(df["thresholdSec"] >= startSec) & (df["thresholdSec"] <= stopSec)]

        """
		if self.stimDict is not None:
			#print(ba.stimDict)
			startSec = self.stimDict['stimStart_sec']
			durSeconds = self.stimDict['stimDur_sec']
			stopSec = startSec + durSeconds
			try:
				df = df[ (df['peak_sec']>=startSec) & (df['peak_sec']<=stopSec)]
			except (KeyError) as e:
				logger.error(e)
		"""
        #
        return df

    def getStimStartStop(self):
        """
        If recorded with a stimGen atf file, get start/stop of stimulus.
        Some sweeps will be black, search for the first sweep that has a stimulus parameters
        assuming the start/stop of each sweep is identical (may not be in the future)

        Returns:
                startSec (float): Start sec of the stimulus
                stopSec (float): Stop sec of the stimulus
        """
        # dfStimFile = self._stimFileAsDataFrame()  # get underlying atf file params (from stimGen2)
        dfStimFile = self._stimFileDf

        if dfStimFile is None:
            # no stim file
            return None, None

        startSec = None
        durSec = None
        stopSec = None
        for idx, sweep in enumerate(range(self.numSweeps)):
            tmpStartSec = dfStimFile.loc[sweep, "start(s)"]
            if not isinstance(tmpStartSec, str):
                startSec = tmpStartSec
                durSec = dfStimFile.loc[sweep, "dur(s)"]
                stopSec = startSec + durSec
                break
        return startSec, stopSec

    def _stimFileAsDataFrame(self):
        """
        If recorded using stimGen atf file.
        Return the stmi parameters as a dataframe
        """
        d = self.stimDict
        if d is None:
            return

        # dList = buildStimDict(d, path=self.filePath)
        # for one in dList:
        # 	print(one)
        df = pd.DataFrame(self._stimDictList)
        return df



from PyQt5 import QtCore, QtWidgets, QtGui

from sanpy.sanpyLogger import get_logger

logger = get_logger(__name__)

import sanpy
from sanpy.interface.plugins import sanpyPlugin


class sanpyLog(sanpyPlugin):
    """
    Plugin to display sanpy.log

    Good example of PyQt plugin
    """

    myHumanName = "SanPy Log"

    def __init__(self, **kwargs):
        """
        Args:
            ba (bAnalysis): Not required
        """
        super(sanpyLog, self).__init__(**kwargs)

        # self.pyqtWindow() # makes self.mainWidget

        layout = QtWidgets.QVBoxLayout()
        widget = QtWidgets.QPlainTextEdit()
        widget.setReadOnly(True)
        layout.addWidget(widget)

        # load sanpy.log
        logFilePath = sanpy.sanpyLogger.getLoggerFile()
        with open(logFilePath, "r") as f:
            lines = f.readlines()
        text = ""
        for line in lines:
            text += line  # line already has CR

        # add text to widget
        widget.document().setPlainText(text)

        logger.info(f"logFilePath: {logFilePath}")

        self._mySetWindowTitle()

        # self.setLayout(layout)
        self.getVBoxLayout().addLayout(layout)


if __name__ == "__main__":
    import sys

    app = QtWidgets.QApplication([])
    spl = sanpyLog()
    spl.show()
    sys.exit(app.exec_())



# 20210610

import os
import sys

# import pathlib
import glob
import importlib
import inspect

from typing import Union, Dict, List, Tuple, Optional

import sanpy

# import sanpy.interface
import sanpy.interface.plugins

from sanpy.sanpyLogger import get_logger

logger = get_logger(__name__)


class bPlugins:
    """Generate a dict of plugins.

    Populate the dict by looking in
        - package, sanpy.interface.plugins
        - folder, <user>/sanpy_plugins
    """

    def __init__(self, sanpyApp: Optional["sanpy.interface.SanPyWindow"] = None):
        self._sanpyApp = sanpyApp

        self.userPluginFolder: str = sanpy._util._getUserPluginFolder()
        if not os.path.isdir(self.userPluginFolder):
            self.userPluginFolder = None
        """path to <user>/plugin_dir"""

        self.pluginDict = {}
        """dict of pointers to open plugins"""

        self._openSet = set()
        """set of open plugins"""

        self.loadPlugins()

    def getSanPyApp(self) -> Optional["sanpy.interface.SanPyWindow"]:
        """Get the underlying SanPy app."""
        return self._sanpyApp

    def loadPlugins(self):
        """Load plugins from both:
         - Package: sanpy.interface.plugins
         - Folder: <user>/sanpy_plugins

        See: sanpy.fileLoaders.fileLoader_base.getFileLoader()
        """
        self.pluginDict = {}

        # Enum is to ignore bPlugins.py class ResponseType(Enum)
        ignoreModuleList = [
            "sanpyPlugin",
            "myWidget",
            "ResponseType",
            "SpikeSelectEvent",
            "basePlotTool",
            "NavigationToolbar2QT",
            "myStatListWidget",
        ]

        #
        # system plugins from sanpy.interface.plugins
        # print('loadPlugins sanpy.interface.plugins:', sanpy.interface.plugins)
        loadedList = []
        for moduleName, obj in inspect.getmembers(sanpy.interface.plugins):
            # print('moduleName:', moduleName, 'obj:', obj)
            if inspect.isclass(obj):
                # logger.info(f'moduleName: {moduleName}')
                if moduleName in ignoreModuleList:
                    # our base plugin class
                    continue
                loadedList.append(moduleName)
                fullModuleName = "sanpy.interface.plugins." + moduleName
                humanName = obj.myHumanName  # myHumanName is a static str
                showInMenu = obj.showInMenu  # showInMenu is a static bool
                pluginDict = {
                    "pluginClass": moduleName,
                    "type": "system",
                    "module": fullModuleName,
                    "path": "",
                    "constructor": obj,
                    "humanName": humanName,
                    "showInMenu": showInMenu,
                }
                if humanName in self.pluginDict.keys():
                    logger.warning(
                        f'Plugin already added "{moduleName}" humanName:"{humanName}"'
                    )
                else:
                    self.pluginDict[humanName] = pluginDict

        # print the loaded plugins
        logger.info(f"Loaded plugins:")
        for loaded in loadedList:
            logger.info(f"    {loaded}")
        # sort
        self.pluginDict = dict(sorted(self.pluginDict.items()))

        #
        # user plugins from files in folder <user>/SanPy/plugins
        loadedModuleList = []
        if self.userPluginFolder is not None:
            files = glob.glob(os.path.join(self.userPluginFolder, "*.py"))
        else:
            files = []

        for file in files:
            if file.endswith("__init__.py"):
                continue

            moduleName = os.path.split(file)[1]
            moduleName = os.path.splitext(moduleName)[0]
            fullModuleName = "sanpy.interface.plugins." + moduleName

            loadedModule = self._module_from_file(fullModuleName, file)

            try:
                oneConstructor = getattr(loadedModule, moduleName)
            except AttributeError as e:
                logger.error(
                    f'Did not load user plugin, make sure file name and class name are the same:"{moduleName}"'
                )
            else:
                humanName = oneConstructor.myHumanName
                showInMenu = oneConstructor.showInMenu
                pluginDict = {
                    "pluginClass": moduleName,
                    "type": "user",
                    "module": fullModuleName,
                    "path": file,
                    "constructor": oneConstructor,
                    "humanName": humanName,
                    "showInMenu": showInMenu,
                }
                if humanName in self.pluginDict.keys():
                    logger.warning(
                        f'Plugin already added "{moduleName}" humanName:"{humanName}"'
                    )
                else:
                    self.pluginDict[humanName] = pluginDict
                    loadedModuleList.append(moduleName)

        # print the loaded plugins
        logger.info(f"Loaded {len(loadedModuleList)} user plugin(s)")
        for loaded in loadedModuleList:
            logger.info(f"    {loaded}")

    def runPlugin(self, pluginName: str, ba: sanpy.bAnalysis, show: bool = True):
        """Run one plugin with given a bAnalysis.

        Args:
            pluginName (str):
            ba (bAnalysis): object
            show:
        """
        if not pluginName in self.pluginDict.keys():
            logger.error(f'Did not find plugin: "{pluginName}"')
            return
        else:
            humanName = self.pluginDict[pluginName]["constructor"].myHumanName

            ltwhTuple = None

            # get the visible x-axis from main app
            startStop = None
            app = self.getSanPyApp()
            _pluginDict = None
            if app is not None:
                startSec = app.startSec
                stopSec = app.stopSec
                if startSec is None and stopSec is None:
                    startStop = None
                else:
                    startStop = [startSec, stopSec]

                # has keys for (l, t, w, h)
                _pluginDict = app.getOptions().preferencesGet("pluginPanels", humanName)
                # ltwhTuple = None
                # if _pluginDict is not None:
                #     ltwhTuple = tuple(
                #         _pluginDict['l'],
                #         _pluginDict['t'],
                #         _pluginDict['w'],
                #         _pluginDict['h'],
                #     )
            logger.info(f"Running plugin:")
            logger.info(f"  pluginName:{pluginName}")
            logger.info(f"  humanName:{humanName}")
            logger.info(f"  startStop:{startStop}")
            if _pluginDict is not None:
                logger.info(f"  _pluginDict:{_pluginDict}")
            logger.info(f"  TODO: put try except back in !!!")
            # TODO: to open PyQt windows, we need to keep a local (persistent) variable
            # try:
            if 1:
                # print(1)
                newPlugin = self.pluginDict[pluginName]["constructor"](
                    ba=ba, bPlugin=self, startStop=startStop
                )
                # print(2)
                if not newPlugin.getInitError() and show:
                    if _pluginDict is not None:
                        newPlugin.setGeometry(
                            _pluginDict["l"],
                            _pluginDict["t"],
                            _pluginDict["w"],
                            _pluginDict["h"],
                        )
                    newPlugin.getWidget().show()
                    # newPlugin.getWidget().raise_()  # bring to front, raise is a python keyword
                    # newPlugin.getWidget().activateWindow()  # bring to front

                # add the plugin to open next time we run

            # except(TypeError) as e:
            #     logger.error(f'Error opening plugin "{pluginName}": {e}')
            #     return
            if not newPlugin.getInitError():
                self._openSet.add(newPlugin)

            return newPlugin

    def getType(self, pluginName):
        """
        returns one of ('system', 'user')
        """
        if not pluginName in self.pluginDict.keys():
            logger.error(f'Did not find plugin: "{pluginName}"')
            return None
        else:
            theRet = self.pluginDict[pluginName]["type"]
            return theRet

    def pluginList(self, forMenu=True):
        """Get list of names of loaded plugins

        Pay attention to showInMenu
        """
        retList = []
        for k, v in self.pluginDict.items():
            # k is the name of the plugin
            # v is a dict with its current state
            # print(' QANON ', k)
            # print('     v:', v)
            if not v["showInMenu"]:
                continue
            retList.append(k)
        return retList

    def getHumanNames(self):
        retList = []
        for k, v in self.pluginDict.items():
            myHumanName = v["myHumanName"]
            retList.append(myHumanName)
        return retList

    def _module_from_file(self, module_name: str, file_path: str):
        """

        Args:
            module_name: Is like sanpy.interface.plugins.onePluginFile
            file_path: Full path to onePluginFile source code (onePluginFile.py)
        """
        spec = importlib.util.spec_from_file_location(module_name, file_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return module

    def slot_closeWindow(self, pluginObj):
        """Close named plugin window.

        Args:
            pluginObj (object): The running plugin object reference.

        Important:
            Need to disconnect signal/slot using _disconnectSignalSlot().
            Mostly connections to main SanPy app signals
        """
        # logger.info(pluginObj)
        try:
            logger.info(f'Removing plugin from _openSet: "{pluginObj.getHumanName()}"')
            # Critical to detatch signal/slot, removing from set does not seem to do this?
            pluginObj._disconnectSignalSlot()
            self._openSet.remove(pluginObj)

            # remove from preferences
            if pluginObj is not None and self.getSanPyApp() is not None:
                self.getSanPyApp().configDict.removePlugin(pluginObj.getHumanName())

        except KeyError as e:
            logger.exception(e)

    def _old_slot_selectSpike(self, sDict):
        """
        On user selection of spike in a plugin

        Tell main app to select a spike everywhere
        """
        logger.info(sDict)
        app = self.getSanPyApp()
        if app is not None:
            spikeNumber = sDict["spikeNumber"]
            doZoom = sDict["doZoom"]
            app.selectSpike(spikeNumber, doZoom=doZoom)

    def _old_slot_selectSpikeList(self, sDict: dict):
        """On user selection of spike(s) in a plugin

        Tell main app to select a spike everywhere

        Args:
            sDict (dict) with keys
                spikeList : List[int]
                doZoom : bool
                ba : sanpy.bAnalysis_.bAnalysis
        """
        logger.info(sDict)
        app = self.getSanPyApp()
        if app is not None:
            spikeList = sDict["spikeList"]
            doZoom = sDict["doZoom"]
            app.selectSpikeList(spikeList, doZoom=doZoom)


def test_print_classes():
    """testing"""
    print("__name__:", __name__)
    for name, obj in inspect.getmembers(sys.modules[__name__]):
        if inspect.isclass(obj):
            print(name, ":", obj)

    print("===")
    for name, obj in inspect.getmembers(sanpy.interface.plugins):
        if inspect.isclass(obj):
            print(name, ":", obj)


if __name__ == "__main__":
    # print_classes()
    # sys.exit(1)

    bp = bPlugins()

    pluginList = bp.pluginList()
    print("pluginList:", pluginList)

    abfPath = "/Users/cudmore/Sites/SanPy/data/19114001.abf"
    ba = sanpy.bAnalysis(abfPath)

    bp.runPlugin("plotRecording", ba)
    # bp.runPlugin('plotRecording3', ba)



import json
import os
from typing import Dict, List, Tuple

import numpy as np  # needed to convert np types to JSON types in save
import pandas as pd
import sanpy

from sanpy.sanpyLogger import get_logger

logger = get_logger(__name__)

defaultVal = float("nan")

# each key in analysisResultDict needs to have the same dict
def getDefaultDict():
    defaultDict = {
        "type": "",  # like: int, float, boolean, list
        "default": "",  # default value, can be 0, None, NaN, ...
        "units": "",  # real world units like point, mV, dvdt
        "depends on detection": "",  # organize documentation and refer to bDetect keys
        "error": "",  # if this analysis results can trigger an error
        "description": "",  # long description for documentation
    }
    return defaultDict.copy()

analysisResultDict = {}

key = "spikeNumber"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "int"
analysisResultDict[key]["default"] = None  # todo: not sure
analysisResultDict[key]["units"] = ""
analysisResultDict[key]["description"] = "Spike number across all sweeps. Zero based."

key = "include"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "bool"
analysisResultDict[key]["default"] = True
analysisResultDict[key]["units"] = ""
analysisResultDict[key][
    "description"
] = "Boolean indication include or not. Can be set by user/programmatically  after analysis."

key = "detectionType"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = ""
analysisResultDict[key]["default"] = None
analysisResultDict[key]["units"] = ""
analysisResultDict[key][
    "description"
] = "Type of detection, either vm or dvdt. See enum sanpy.bDetection.detectionTypes"

#
# start real analysis results
key = "sweep"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "int"
analysisResultDict[key]["default"] = 0  # todo: not sure
analysisResultDict[key]["units"] = ""
analysisResultDict[key]["description"] = "Sweep number of analyzed sweep. Zero based."

key = "epoch"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "int"
analysisResultDict[key]["default"] = np.nan  # todo: not sure
analysisResultDict[key]["units"] = ""
analysisResultDict[key][
    "description"
] = "Stimulus epoch number the spike occured in. Zero based."

key = "epochLevel"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "float"
analysisResultDict[key]["default"] = np.nan  # todo: not sure
analysisResultDict[key]["units"] = ""
analysisResultDict[key]["description"] = "Epoch level (DAC) stimulus during the spike."

key = "sweepSpikeNumber"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "int"
analysisResultDict[key]["default"] = None  # todo: not sure
analysisResultDict[key]["units"] = ""
analysisResultDict[key]["description"] = "Spike number within the sweep. Zero based."

key = "userType"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "int"
analysisResultDict[key]["default"] = 0
analysisResultDict[key]["units"] = ""
analysisResultDict[key][
    "description"
] = "Integer indication user type. Can be set by user/programmatically  after analysis."

key = "errors"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "list"
analysisResultDict[key]["default"] = []
analysisResultDict[key]["units"] = ""
analysisResultDict[key][
    "description"
] = "List of dictionary to hold detection errors for this spike"

key = "analysisDate"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "str"
analysisResultDict[key]["default"] = ""
analysisResultDict[key]["units"] = ""
analysisResultDict[key]["description"] = "Date of analysis in yyyymmdd format."

key = "analysisTime"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "str"
analysisResultDict[key]["default"] = ""
analysisResultDict[key]["units"] = ""
analysisResultDict[key]["description"] = "Time of analysis in hh:mm:ss 24 hours format."

key = "modDate"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "str"
analysisResultDict[key]["default"] = ""
analysisResultDict[key]["units"] = ""
analysisResultDict[key][
    "description"
] = "Modification date if AP is modified after detection."

key = "modTime"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "str"
analysisResultDict[key]["default"] = ""
analysisResultDict[key]["units"] = ""
analysisResultDict[key][
    "description"
] = "Modification time if AP is modified after detection."

key = "analysisVersion"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "str"
analysisResultDict[key]["default"] = ""
analysisResultDict[key]["units"] = ""
analysisResultDict[key][
    "description"
] = "Analysis version when analysis was run. See sanpy.analysisVersion"

key = "interfaceVersion"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "str"
analysisResultDict[key]["default"] = ""
analysisResultDict[key]["units"] = ""
analysisResultDict[key][
    "description"
] = "Interface version string when analysis was run. See sanpy.interfaceVersion"

key = "file"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "str"
analysisResultDict[key]["default"] = ""
analysisResultDict[key]["units"] = ""
analysisResultDict[key]["description"] = "Name of raw data file analyzed"

key = "cellType"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "str"
analysisResultDict[key]["default"] = ""
analysisResultDict[key]["units"] = ""
analysisResultDict[key]["description"] = "User specified cell type"

key = "sex"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "str"
analysisResultDict[key]["default"] = ""
analysisResultDict[key]["units"] = ""
analysisResultDict[key]["description"] = "User specified sex"

key = "condition"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "str"
analysisResultDict[key]["default"] = ""
analysisResultDict[key]["units"] = ""
analysisResultDict[key]["description"] = "User specified condition"

#
# detection parameters
key = "dvdtThreshold"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "float"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "dvdt"
analysisResultDict[key]["depends on detection"] = "dvdtThreshold"
analysisResultDict[key]["description"] = "AP Threshold in derivative dv/dt"

key = "mvThreshold"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "float"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "mV"
analysisResultDict[key]["depends on detection"] = "mvThreshold"
analysisResultDict[key]["description"] = "AP Threshold in primary recording mV"

# todo: what about solay ...
key = "medianFilter"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "int"
analysisResultDict[key]["default"] = 0
analysisResultDict[key]["units"] = ""
analysisResultDict[key]["depends on detection"] = "medianFilter"
analysisResultDict[key][
    "description"
] = "Median filter to generate filtered vm and dvdt. Value 0 indicates no filter."

key = "halfHeights"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "list"
analysisResultDict[key]["default"] = []
analysisResultDict[key]["units"] = ""
analysisResultDict[key]["depends on detection"] = "halfHeights"
analysisResultDict[key][
    "description"
] = "List of int to specify half-heights like [10, 20, 50, 80, 90]."

#
# actual value coming out of detection !!!!
key = "thresholdPnt"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "int"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "point"
analysisResultDict[key]["description"] = "AP threshold point"

key = "thresholdSec"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "float"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "sec"
analysisResultDict[key]["description"] = "AP threshold seconds"

key = "thresholdVal"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "float"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "mV"  # TODO: will be pA for voltage-clamp
analysisResultDict[key]["description"] = "Value of Vm at AP threshold point."

key = "thresholdVal_dvdt"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "float"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "dvdt"  # TODO: will be pA for voltage-clamp
analysisResultDict[key]["description"] = "Value of dvdt at AP threshold point."

key = "dacCommand"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "float"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "mV"  # TODO: will be pA for voltage-clamp
analysisResultDict[key]["description"] = "Value of DAC command at AP threshold point."

key = "peakPnt"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "int"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "point"
analysisResultDict[key]["depends on detection"] = "(onlyPeaksAbove_mV, peakWindow_ms)"
analysisResultDict[key]["description"] = "AP peak point."

key = "peakSec"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "float"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "sec"
analysisResultDict[key]["description"] = "AP peak seconds."

key = "peakVal"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "float"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "mV"  # voltage-clamp'
analysisResultDict[key]["description"] = "Value of Vm at AP peak point."

key = "peakHeight"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "float"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "mV"  # voltage-clamp
analysisResultDict[key][
    "description"
] = "Difference between peakVal minus thresholdVal."

key = "timeToPeak_ms"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "float"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "ms"  # voltage-clamp'
analysisResultDict[key]["description"] = "Time to peak (ms) after TOP."

key = "preMinPnt"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "int"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "point"  # voltage-clamp
analysisResultDict[key]["depends on detection"] = "mdp_ms"
analysisResultDict[key][
    "description"
] = "Minimum before an AP taken from predefined window."

key = "preMinVal"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "float"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "mV"  # voltage-clamp
analysisResultDict[key][
    "description"
] = "Minimum before an AP taken from predefined window."

key = "preLinearFitPnt0"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "int"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "point"  # voltage-clamp
analysisResultDict[key][
    "description"
] = "Point where pre linear fit starts. Used for EDD Rate"

key = "preLinearFitPnt1"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "int"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "point"  # voltage-clamp
analysisResultDict[key][
    "description"
] = "Point where pre linear fit stops. Used for EDD Rate"

key = "earlyDiastolicDuration_ms"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "float"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "ms"  # voltage-clamp
analysisResultDict[key]["description"] = "Time (ms) between start/stop of EDD."

key = "preLinearFitVal0"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "float"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "mv"  # voltage-clamp
analysisResultDict[key]["description"] = ""

key = "preLinearFitVal1"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "float"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "mv"
analysisResultDict[key]["description"] = ""

key = "earlyDiastolicDurationRate"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "float"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "mv/S"
analysisResultDict[key][
    "description"
] = "Early diastolic duration rate, the slope of the linear fit between start/stop of EDD."

key = "lateDiastolicDuration"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "float"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = ""
analysisResultDict[key]["description"] = "Depreciated"

key = "preSpike_dvdt_max_pnt"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "int"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "point"
analysisResultDict[key][
    "description"
] = "Point corresponding to peak in dv/dt before an AP."

key = "preSpike_dvdt_max_val"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "float"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "mV"
analysisResultDict[key]["description"] = "Value of Vm at peak of dv/dt before an AP."

key = "preSpike_dvdt_max_val2"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "float"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "dv/dt"
analysisResultDict[key]["description"] = "Value of dv/dt at peak of dv/dt before an AP."

key = "postSpike_dvdt_min_pnt"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "int"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "point"
analysisResultDict[key]["depends on detection"] = "dvdtPostWindow_ms"
analysisResultDict[key][
    "description"
] = "Point corresponding to min in dv/dt after an AP."

key = "postSpike_dvdt_min_val"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "float"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "mV"
analysisResultDict[key]["description"] = "Value of Vm at minimum of dv/dt after an AP."

key = "postSpike_dvdt_min_val2"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "float"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "dvdt"
analysisResultDict[key][
    "description"
] = "Value of dv/dt at minimum of dv/dt after an AP."

key = "isi_pnts"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "int"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "point"
analysisResultDict[key]["depends on detection"] = "refractory_ms"
analysisResultDict[key][
    "description"
] = "Inter-Spike-Interval (points) with respect to previous AP."

key = "isi_ms"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "float"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "ms"
analysisResultDict[key][
    "description"
] = "Inter-Spike-Interval (ms) with respect to previous AP."

key = "spikeFreq_hz"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "float"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "Hz"
analysisResultDict[key]["description"] = "AP frequency with respect to previous AP."

key = "cycleLength_pnts"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "int"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "point"
analysisResultDict[key][
    "description"
] = "Points between APs with respect to previous AP."

key = "cycleLength_ms"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "int"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "point"
analysisResultDict[key][
    "description"
] = "Time (ms) between APs with respect to previous AP."

key = "diastolicDuration_ms"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "float"
analysisResultDict[key]["default"] = defaultVal
analysisResultDict[key]["units"] = "ms"
analysisResultDict[key][
    "description"
] = "Time (ms) between minimum before AP (preMinPnt) and AP time (thresholdPnt)."

key = "widths"
analysisResultDict[key] = getDefaultDict()
analysisResultDict[key]["type"] = "list"
analysisResultDict[key]["default"] = []
analysisResultDict[key]["units"] = ""
analysisResultDict[key][
    "description"
] = "A list of dict to hold half-height information for each half-height in detection halfHeights."

# need to define keys width_<i> from halfHeights list [10, 20, 50, 80, 90]
for i in [10, 20, 50, 80, 90]:
    key = "widths_" + str(i)
    analysisResultDict[key] = getDefaultDict()
    analysisResultDict[key]["type"] = "int"
    analysisResultDict[key]["default"] = defaultVal
    analysisResultDict[key]["units"] = "percent"
    analysisResultDict[key]["depends on detection"] = "halfWidthWindow_ms"
    analysisResultDict[key]["description"] = f"Width (ms) at half-height {i} %."

# now we have (AP threshold, AP peak), derive stats about each AP


def printDocs():
    """
    Print out human readable detection parameters and convert to markdown table.

    Requires:
        pip install tabulate

    See: bDetection.printDocs()
    """
    import pandas as pd
    from datetime import datetime

    logger.info("Ensure there are no errors")

    dictList = []
    for k, v in analysisResultDict.items():
        # iterating on getDefaultDict() to ensure all code above has valid k/v pairs
        # lineStr = k + '\t'
        oneDict = {
            "Name": k,
        }
        for k2 in getDefaultDict().keys():
            # print(f'  {k}: {k2}: {v[k2]}')
            # lineStr += f'{v[k2]}' + '\t'
            oneDict[k2] = v[k2]
        #
        # print(lineStr)

        dictList.append(oneDict)

        # check that k is in headerDefaultDict
        for k3 in v:
            if not k3 in getDefaultDict().keys():
                logger.error(f'Found extra key "{k}" in "analysisResultDict"')

    #
    df = pd.DataFrame(dictList)

    if 1:
        # to markdown for mkdocs md file
        # str = df.to_markdown()
        str = df.to_html()
        myDate = datetime.today().strftime("%Y-%m-%d")
        print(f"Generated {myDate} with sanpy.analysisVersion {sanpy.analysisVersion}")
        print(str)

    if 0:
        path = "/Users/cudmore/Desktop/sanpy-analysis-results.csv"
        print("saving:", path)
        df.to_csv(path, index=False)


class NumpyEncoder(json.JSONEncoder):
    """Special json encoder for numpy types"""

    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


class analysisResultList:
    """Class encapsulating a list of analysis results.
    
    Each row is an analysisResultDict for one spike.

    These are keys in bAnalysis_ spike dict and columns in output reports
    """

    def __init__(self):
        # one copy for entire list

        # TODO: put xxx in a function getAnalysisResltDict()
        self._dDict = analysisResultDict

        # list of analysisResultDict
        self._myList = []

        self._iterIdx = -1

    def setFromListDict(self, listOfDict: List[dict]):
        """Set analysis results from a list of dict.

        Used when loading sanpy.bAnalysis from h5 file.

        When we create self (during spike detect) we have a list of class analysisResult.
        When we save/load we have a list of dict.

        This is assuming we re-create self every time we do spike detection
        """

        # do not do this, we are an analysisResultList as a list of analysisResult
        # self._myList = listOfDict

        self._myList = []
        for oneDict in listOfDict:
            oneAnalysisResult = analysisResult(theDict=oneDict)
            self._myList.append(oneAnalysisResult)

    def analysisDate(self):
        if len(self) > 0:
            return self._myList[0]["analysisDate"]
        else:
            return None

    def analysisTime(self):
        if len(self) > 0:
            return self._myList[0]["analysisTime"]
        else:
            return None

    def _old_save(self, saveBase):
        savePath = saveBase + "-analysis.json"

        analysisList = self.asList()

        # print(analysisList[0].print())
        print(self._myList[0])

        with open(savePath, "w") as f:
            json.dump(analysisList, f, cls=NumpyEncoder, indent=4)

    def _old_load(self, loadBase):
        loadPath = loadBase + "-analysis.json"

        if not os.path.isfile(loadPath):
            logger.error(f"Did not find file: {loadPath}")
            return

        with open(loadPath, "r") as f:
            self._myList = json.load(f)

    def appendDefault(self):
        """Append a spike to analysis.

        Used in bAnalysis spike detection.
        """
        oneResult = analysisResult()
        self._myList.append(oneResult)

    def appendAnalysis(self, analysisResultList):
        for analysisResult in analysisResultList:
            # analysisResult is for one spike
            self._myList.append(analysisResult)

    def addAnalysisResult(self, theKey, theDefault=None):
        # go through list and add to each [i] dict
        for spike in self:
            spike.addNewKey(theKey, theDefault=theDefault)

    def asList(self):
        """
        Return underlying list.
        """
        # return [spike.asDict() for spike in self._myList]
        return [x.asDict() for x in self._myList]

    def asDataFrame(self):
        """
        Note: underlying _myList is a list of analysisResult
        """
        return pd.DataFrame(self.asList())

    def __getitem__(self, key):
        """
        Allow [] indexing with self[int].
        """
        try:
            # return self._dDict[key]['currentValue']
            return self._myList[key]
        except IndexError as e:
            logger.error(f"{e}")

    def __len__(self):
        """Allow len() with len(this)"""
        return len(self._myList)

    def __iter__(self):
        """Allow iteration with "for item in self"
        """
        _iterIdx = -1
        return self

    def __next__(self):
        """Allow iteration with "for item in self"
        """
        self._iterIdx += 1
        if self._iterIdx >= len(self._myList):
            self._iterIdx = -1  # reset to initial value
            raise StopIteration
        else:
            return self._myList[self._iterIdx]


class analysisResult:
    def __init__(self, theDict=None):
        """Create an anlysis item (for one spike)

        Args:
            theDict: Pre-existing dict when we load form h5 file
        """
        # this is the raw definition of analysis results (See above)
        # pull from this to create key/value self.rDict
        # self._dDict = analysisResultDict
        defaultDict = analysisResultDict
        # this is simple key/value pairs as we will in detection
        self._rDict = {}
        for k, v in defaultDict.items():
            default = v["default"]
            self._rDict[k] = default

        if theDict is not None:
            for k, v in theDict.items():
                self[k] = v  # calls __setitem__()

    # this was interfering with converting to DataFrame ???
    """
    def __str__(self):
        printList = []
        for k,v in self._rDict.items():
            if isinstance(v, dict):
                for k2,v2 in v.items():
                    printList.append(f'  {k2} : {v2} {type(v2)}')
            else:
                printList.append(f'{k} : {v} {type(v)}')
        return '\n'.join(printList)
    """

    def print(self):
        printList = []
        for k, v in self._rDict.items():
            if isinstance(v, list):
                for item in v:
                    for k2, v2 in item.items():
                        printList.append(f"  {k2} : {v2} {type(v2)}")
            else:
                printList.append(f"{k} : {v} {type(v)}")
        return "\n".join(printList)

    def addNewKey(self, theKey, theDefault=None):
        """
        Add a new key to this spike.

        Returns: (bool) True if new key added, false if key already exists.
        """
        if theDefault is None:
            # theType = 'float'
            theDefault = float("nan")

        # check if key exists
        keyExists = theKey in self._rDict.keys()
        addedKey = False
        if keyExists:
            # key exists, don't modify
            # logger.warning(f'The key "{theKey}" already exists and has value "{self._rDict[theKey]}"')
            pass
        else:
            self._rDict[theKey] = theDefault
            addedKey = True

        #
        return addedKey

    def asDict(self):
        """
        Returns underlying dictionary
        """
        return self._rDict

    def __getitem__(self, key):
        # to mimic a dictionary
        ret = None
        try:
            # return self._dDict[key]['currentValue']
            ret = self._rDict[key]
        except KeyError as e:
            logger.error(f'Error getting key "{key}" exception:{e}')
            raise
        #
        return ret

    def __setitem__(self, key, value):
        # to mimic a dictionary
        try:
            # self._dDict[key]['currentValue'] = value
            self._rDict[key] = value
        except KeyError as e:
            logger.error(f"{e}")

    def items(self):
        # to mimic a dictionary
        return self._rDict.items()

    def keys(self):
        # to mimic a dictionary
        return self._rDict.keys()


def test():
    ar = analysisResultList()

    for k, v in ar.items():
        print(k, v)

    key = "analysisVersion"
    print(f'key:{key} value:"{ar[key]}" type:{type(ar[key])}')

    key = "errors"
    print(f'key:{key} value:"{ar[key]}" type:{type(ar[key])}')


def test2():
    # load abf
    path = "data/19114000.abf"
    ba = sanpy.bAnalysis(path)

    # detect
    sweepNumber = 0
    detectionClass = ba.detectionClass
    detectionClass["verbose"] = True
    ba.spikeDetect2__(sweepNumber, detectionClass)

    """
    printSpikeNum = 4
    print(f'== printing spike {printSpikeNum}')
    ba.printSpike(printSpikeNum)
    """

    # ba.printErrors()

    sd = ba.getSpikeDictionaries()
    """
    for idx, s in enumerate(sd):
        if idx == 2:
            print(s)
    """


if __name__ == "__main__":
    # test()
    # test2()
    printDocs()



import os
from collections import OrderedDict
from pprint import pprint
from typing import Union, Dict, List, Tuple, Optional

import numpy as np
import pandas as pd

import sanpy

from sanpy.sanpyLogger import get_logger

logger = get_logger(__name__)


class bExport:
    """Once analysis is performed with sanpy.bAnalysis.spikeDetect(),
        reports can be generated with the bExport class.

    Example reports are:

    - Generating reports as a Pandas DataFrame.
    - (depreciated) Saving reports as a Microsoft Excel file.
    - Saving reports as a CSV text files.
    """

    def __init__(self, ba):
        """
        Args:
            ba (sanpy.bAnalysis): A bAnalysis object that has had spikes detected with detectSpikes().
        """
        self.ba = ba
        self.sweepNumber = 0

    def old_report(self, theMin, theMax):
        """
        Get entire spikeDict as a Pandas DataFrame.

        Args:
            theMin (float): Start seconds of the analysis
            theMax (float): Stop seconds of the analysis

        Returns:
            df: Pandas DataFrame
        """
        if theMin is None or theMax is None:
            # return None
            theMin = 0
            theMax = self.ba.fileLoader.recordingDur

        logger.info(f"theMin:{theMin} theMax:{theMax}")

        df = self.ba.asDataFrame()
        df = df[(df["thresholdSec"] >= theMin) & (df["thresholdSec"] <= theMax)]

        # added when trying to make scatterwidget for one file
        df["Condition"] = ""  # df['condition1']
        df["File Number"] = ""  # df['condition2']
        df["Sex"] = ""  # df['condition3']
        df["Region"] = ""  # df['condition4']

        # make new column with sex/region encoded
        """
        tmpNewCol = 'RegSex'
        self.ba.masterDf[tmpNewCol] = ''
        for tmpRegion in ['Superior', 'Inferior']:
            for tmpSex in ['Male', 'Female']:
                newEncoding = tmpRegion[0] + tmpSex[0]
                regSex = self.ba.masterDf[ (self.ba.masterDf['Region']==tmpRegion) & (self.ba.masterDf['Sex']==tmpSex)]
                regSex = (self.ba.masterDf['Region']==tmpRegion) & (self.ba.masterDf['Sex']==tmpSex)
                print('newEncoding:', newEncoding, 'regSex:', regSex.shape)
                self.ba.masterDf.loc[regSex, tmpNewCol] = newEncoding
        """

        # want this but region/sex/condition are not defined
        # print('bExport.report()')
        # print(df.head())
        tmpNewCol = "CellTypeSex"
        cellTypeStr = df["cellType"].iloc[0]
        sexStr = df["sex"].iloc[0]
        # print('cellTypeStr:', cellTypeStr, 'sexStr:', sexStr)
        regSexEncoding = cellTypeStr + sexStr
        df[tmpNewCol] = regSexEncoding

        minStr = "%.2f" % (theMin)
        maxStr = "%.2f" % (theMax)
        minStr = minStr.replace(".", "_")
        maxStr = maxStr.replace(".", "_")

        # TODO: bytestreams are not strictly from a hdd folder or file
        fileName = self.ba.fileLoader.filename
        if fileName is not None:
            fileName, tmpExt = os.path.splitext(fileName)
            analysisName = fileName + "_s" + minStr + "_s" + maxStr
            # print('    minStr:', minStr, 'maxStr:', maxStr, 'analysisName:', analysisName)
        else:
            analysisName = "bytestream"
        df["analysisname"] = analysisName

        return df

    def report3(self, sweep='All',
                epoch='All',
                theMin : Optional[float] = None,
                theMax : Optional[float] = None,
                ) -> pd.DataFrame:
        """Generate a full report of all spike columns.
        
        Like what is save in csv but limited by sweep, epoch, etc
        """

        if self.ba.numSpikes == 0:
            logger.warning(f"did not find and spikes for summary")
            return None

        df = self.ba.asDataFrame()  # full df with all spikes

        spikeList = self.ba.getStat('spikeNumber', sweepNumber=sweep, epochNumber=epoch)

        # reduce to spikes in list
        df = df.loc[df['spikeNumber'].isin(spikeList)]

        if theMin is not None and theMax is not None:
            df = df[ (df['thresholdSec']>=theMin) & (df['thresholdSec']<theMax)]
        
        return df
    
    def report2(self, sweep='All',
                epoch='All',
                theMin : Optional[float] = None,
                theMax : Optional[float] = None,
                ) -> pd.DataFrame:

        """Generate a human readable report of spikes.
        Include spike times between theMin and theMax (Sec).

        Args:
            sweep ('All' or int') : 'All' for all sweeps or int for one sweep
            epoch ('All' or int) : 'All' for all epochs or int for one epoch
            theMin (float): Start seconds to save, inclusive
            theMax (float): Stop seconds to save, inclusive

        Returns:
            df: pd.DataFrame
        """

        spikeList = self.ba.getStat('spikeNumber', sweepNumber=sweep, epochNumber=epoch)

        newList = []
        for spikeIdx in spikeList:
            spike = self.ba.getOneSpikeDict(spikeIdx)
            
            # if current spike time is out of bounds
            # then continue (e.g. it is not between theMin (sec) and theMax (sec)
            spikeTime_sec = self.ba.fileLoader.pnt2Sec_(spike["thresholdPnt"])
            if theMin is not None and theMax is not None:
                if spikeTime_sec < theMin or spikeTime_sec > theMax:
                    continue

            spikeDict = (
                OrderedDict()
            )  # use OrderedDict so Pandas output is in the correct order

            spikeDict["Spike"] = spikeIdx
            spikeDict["Take Off Potential (s)"] = self.ba.fileLoader.pnt2Sec_(
                spike["thresholdPnt"]
            )
            spikeDict["Take Off Potential (ms)"] = self.ba.fileLoader.pnt2Ms_(
                spike["thresholdPnt"]
            )
            spikeDict["Take Off Potential (mV)"] = spike["thresholdVal"]
            spikeDict["AP Peak (ms)"] = self.ba.fileLoader.pnt2Ms_(spike["peakPnt"])
            spikeDict["AP Peak (mV)"] = spike["peakVal"]
            spikeDict["AP Height (mV)"] = spike["peakHeight"]
            spikeDict["Pre AP Min (mV)"] = spike["preMinVal"]
            # spikeDict['Post AP Min (mV)'] = spike['postMinVal']
            #
            # spikeDict['AP Duration (ms)'] = spike['apDuration_ms']
            spikeDict["Early Diastolic Duration (ms)"] = spike[
                "earlyDiastolicDuration_ms"
            ]
            spikeDict["Early Diastolic Depolarization Rate (dV/s)"] = spike[
                "earlyDiastolicDurationRate"
            ]  # abb 202012
            spikeDict["Diastolic Duration (ms)"] = spike["diastolicDuration_ms"]
            #
            spikeDict["Inter-Spike-Interval (ms)"] = spike["isi_ms"]
            spikeDict["Spike Frequency (Hz)"] = spike["spikeFreq_hz"]
            spikeDict["ISI (ms)"] = spike["isi_ms"]

            spikeDict["Cycle Length (ms)"] = spike["cycleLength_ms"]

            spikeDict["Max AP Upstroke (dV/dt)"] = spike["preSpike_dvdt_max_val2"]
            spikeDict["Max AP Upstroke (mV)"] = spike["preSpike_dvdt_max_val"]

            spikeDict["Max AP Repolarization (dV/dt)"] = spike[
                "postSpike_dvdt_min_val2"
            ]
            spikeDict["Max AP Repolarization (mV)"] = spike["postSpike_dvdt_min_val"]

            # half-width
            for widthDict in spike["widths"]:
                keyName = "width_" + str(widthDict["halfHeight"])
                spikeDict[keyName] = widthDict["widthMs"]

            spikeDict["File"] = self.ba.fileLoader.filename

            # errors
            # spikeDict['numError'] = spike['numError']
            spikeDict["errors"] = spike["errors"]

            # append
            newList.append(spikeDict)

        df = pd.DataFrame(newList)
        return df

    def getSummary(self,
                   sweep='All',
                    epoch='All',
                    theMin: float = None, 
                    theMax: float = None
                    ) -> pd.DataFrame:
        """Get analysis summary as df.
        
        This adds some header information to spike report bExport.report2().
        """

        if self.ba.numSpikes == 0:
            logger.warning(f"did not find and spikes for summary")
            return None

        # if theMin is None or theMax is None:
        #     theMin = 0
        #     theMax = self.ba.fileLoader.recordingDur

        #
        # cardiac style analysis to sheet 'cardiac'
        # human readable columns
        cardiac_df = self.report2(sweep=sweep,
                                  epoch=epoch,
                                  theMin=theMin,
                                  theMax=theMax)

        dDict = self.ba.getDetectionDict()

        #
        # header sheet
        headerDict = OrderedDict()
        filePath, fileName = os.path.split(self.ba.fileLoader.filepath)
        headerDict["File Name"] = [fileName]
        headerDict["File Path"] = [filePath]

        headerDict["Cell Type"] = [dDict["cellType"]]
        headerDict["Sex"] = [dDict["sex"]]
        headerDict["Condition"] = [dDict["condition"]]

        headerDict["Date Analyzed"] = [
            self.ba.analysisDate
        ]  # pulled from first detected spike
        headerDict["Time Analyzed"] = [self.ba.analysisTime]

        headerDict["Detection Type"] = [dDict["detectionType"]]
        headerDict["dV/dt Threshold"] = [dDict["dvdtThreshold"]]
        # headerDict['mV Threshold'] = [self.ba.mvThreshold] # abb 202012
        headerDict["Vm Threshold (mV)"] = [dDict["mvThreshold"]]
        # headerDict['Median Filter (pnts)'] = [self.ba.medianFilter]
        headerDict["Analysis Version"] = [sanpy.analysisVersion]
        headerDict["Interface Version"] = [sanpy.interfaceVersion]

        # headerDict['Analysis Start (sec)'] = [self.ba.startSeconds]
        # headerDict['Analysis Stop (sec)'] = [self.ba.stopSeconds]
        headerDict["Sweep Number"] = ["Default 0"]  # [self.ba.currentSweep]
        headerDict["Number of Sweeps"] = [self.ba.fileLoader.numSweeps]
        headerDict["Export Start (sec)"] = [
            float("%.2f" % (theMin))
        ]  # on export, x-axis of raw plot will be ouput
        headerDict["Export Stop (sec)"] = [
            float("%.2f" % (theMax))
        ]  # on export, x-axis of raw plot will be ouput

        # 'stats' has xxx columns (name, mean, sd, se, n)
        headerDict["stats"] = []

        ignoreColumns = ["Spike", "File"]
        for idx, col in enumerate(cardiac_df):
            if col in ignoreColumns:
                # in general, skip non numerical columns
                continue
            headerDict[col] = []

        # mean
        theMean = cardiac_df.mean(numeric_only=True)  # skipna default is True

        logger.info('cardiac_df:')
        print(cardiac_df)
        logger.info('theMean:')
        print(theMean)

        theMean["errors"] = ""
                
        # sd
        theSD = cardiac_df.std(numeric_only=True)  # skipna default is True
        theSD["errors"] = ""
        # se
        theSE = cardiac_df.sem(numeric_only=True)  # skipna default is True
        theSE["errors"] = ""
        # n
        theN = cardiac_df.count(numeric_only=True)  # skipna default is True
        theN["errors"] = ""

        statCols = ["mean", "sd", "se", "n"]
        for j, stat in enumerate(statCols):
            if j == 0:
                pass
            else:
                # need to append columns to keep Excel sheet columns in sync
                # for k,v in headerDict.items():
                #    headerDict[k].append('')

                headerDict["File Name"].append("")
                headerDict["File Path"].append("")
                headerDict["Cell Type"].append("")
                headerDict["Sex"].append("")
                headerDict["Condition"].append("")
                #
                headerDict["Date Analyzed"].append("")
                headerDict["Time Analyzed"].append("")
                headerDict["Detection Type"].append("")
                headerDict["dV/dt Threshold"].append("")
                headerDict["Vm Threshold (mV)"].append("")
                # headerDict['Median Filter (pnts)'].append('')
                headerDict["Analysis Version"].append("")
                headerDict["Interface Version"].append("")
                headerDict["Sweep Number"].append("")
                headerDict["Number of Sweeps"].append("")
                headerDict["Export Start (sec)"].append("")
                headerDict["Export Stop (sec)"].append("")

            # a dictionary key for each stat
            headerDict["stats"].append(stat)
            for idx, col in enumerate(cardiac_df):
                if col in ignoreColumns:
                    # in general, need to ignore string columns
                    # headerDict[col].append('')
                    continue
                # headerDict[col].append('')
                if stat == "mean":
                    headerDict[col].append(theMean[col])
                elif stat == "sd":
                    headerDict[col].append(theSD[col])
                elif stat == "se":
                    headerDict[col].append(theSE[col])
                elif stat == "n":
                    headerDict[col].append(theN[col])

        # end for j, stat
        # print('=== headerDict')
        # for k,v in headerDict.items():
        #    print(k, ':', v)

        # dict to pandas dataframe
        df = pd.DataFrame(headerDict).T
        df.insert(0, "", headerDict.keys(), allow_duplicates=True)

        return df

    def saveReport(
        self,
        savefile,
        theMin=None,
        theMax=None,
        saveExcel=True,
        alsoSaveTxt=True,
        verbose=True,
    ):
        """
        Save a spike report for detected spikes between theMin (sec) and theMax (sec).

        This is used by main interface 'Export Spike Report'

        Args:
            savefile (str): path to xlsx file
            theMin (float): start/stop seconds of the analysis
            theMax (float): start/stop seconds of the analysis
            saveExcel (bool):
            alsoSaveTxt (bool):

        Return:
            str: analysisName
            df: df
        """
        if theMin is None or theMax is None:
            theMin = 0
            theMax = self.ba.fileLoader.recordingDur

        # always grab a df to the entire analysis (not sure what I will do with this)
        # df = self.ba.report() # report() is my own 'bob' verbiage

        theRet = None

        logger.warning("NEVER SAVING EXCEL !!! dec 2022")
        saveExcel = False
        if saveExcel and savefile:
            # if verbose: print('    bExport.saveReport() saving user specified .xlsx file:', savefile)
            excelFilePath = savefile
            writer = pd.ExcelWriter(excelFilePath, engine="xlsxwriter")

            #
            # cardiac style analysis to sheet 'cardiac'
            cardiac_df = self.report2(theMin, theMax)  # report2 is more 'cardiac'

            dDict = self.ba.getDetectionDict()
            dateAnalyzed = self.ba.dateAnalyzed
            timeAnalyzed = self.ba.dateAnalyzed

            #
            # header sheet
            headerDict = OrderedDict()
            filePath, fileName = os.path.split(self.ba.filepath)
            headerDict["File Name"] = [fileName]
            headerDict["File Path"] = [filePath]

            headerDict["Cell Type"] = [dDict["cellType"]]
            headerDict["Sex"] = [dDict["sex"]]
            headerDict["Condition"] = [dDict["condition"]]

            # todo: get these params in ONE dict inside self.ba
            # dateAnalyzed, timeAnalyzed = self.ba.dateAnalyzed.split(' ')
            headerDict["Date Analyzed"] = [dateAnalyzed]
            headerDict["Time Analyzed"] = [timeAnalyzed]
            headerDict["Detection Type"] = [dDict["detectionType"]]
            headerDict["dV/dt Threshold"] = [dDict["dvdtThreshold"]]
            # headerDict['mV Threshold'] = [self.ba.mvThreshold] # abb 202012
            headerDict["Vm Threshold (mV)"] = [dDict["mvThreshold"]]
            # headerDict['Median Filter (pnts)'] = [self.ba.medianFilter]
            headerDict["Analysis Version"] = [sanpy.analysisVersion]
            headerDict["Interface Version"] = [sanpy.interfaceVersion]

            # headerDict['Analysis Start (sec)'] = [self.ba.startSeconds]
            # headerDict['Analysis Stop (sec)'] = [self.ba.stopSeconds]
            headerDict["Sweep Number"] = ["Default 0"]  # [self.ba.currentSweep]
            headerDict["Number of Sweeps"] = [self.ba.fileLoader.numSweeps]
            headerDict["Export Start (sec)"] = [
                float("%.2f" % (theMin))
            ]  # on export, x-axis of raw plot will be ouput
            headerDict["Export Stop (sec)"] = [
                float("%.2f" % (theMax))
            ]  # on export, x-axis of raw plot will be ouput

            # 'stats' has xxx columns (name, mean, sd, se, n)
            headerDict["stats"] = []

            ignoreColumns = ["Spike", "File"]
            for idx, col in enumerate(cardiac_df):
                if col in ignoreColumns:
                    # in general, need to ignore string columns
                    # headerDict[col].append('')
                    continue
                headerDict[col] = []

            # mean
            theMean = cardiac_df.mean()  # skipna default is True
            theMean["errors"] = ""
            # sd
            theSD = cardiac_df.std()  # skipna default is True
            theSD["errors"] = ""
            # se
            theSE = cardiac_df.sem()  # skipna default is True
            theSE["errors"] = ""
            # n
            theN = cardiac_df.count()  # skipna default is True
            theN["errors"] = ""

            statCols = ["mean", "sd", "se", "n"]
            for j, stat in enumerate(statCols):
                if j == 0:
                    pass
                else:
                    # need to append columns to keep Excel sheet columns in sync
                    # for k,v in headerDict.items():
                    #    headerDict[k].append('')

                    headerDict["File Name"].append("")
                    headerDict["File Path"].append("")
                    headerDict["Cell Type"].append("")
                    headerDict["Sex"].append("")
                    headerDict["Condition"].append("")
                    #
                    headerDict["Date Analyzed"].append("")
                    headerDict["Time Analyzed"].append("")
                    headerDict["Detection Type"].append("")
                    headerDict["dV/dt Threshold"].append("")
                    headerDict["Vm Threshold (mV)"].append("")
                    # headerDict['Median Filter (pnts)'].append('')
                    headerDict["Analysis Version"].append("")
                    headerDict["Interface Version"].append("")
                    headerDict["Sweep Number"].append("")
                    headerDict["Number of Sweeps"].append("")
                    headerDict["Export Start (sec)"].append("")
                    headerDict["Export Stop (sec)"].append("")

                # a dictionary key for each stat
                headerDict["stats"].append(stat)
                for idx, col in enumerate(cardiac_df):
                    if col in ignoreColumns:
                        # in general, need to ignore string columns
                        # headerDict[col].append('')
                        continue
                    # headerDict[col].append('')
                    if stat == "mean":
                        headerDict[col].append(theMean[col])
                    elif stat == "sd":
                        headerDict[col].append(theSD[col])
                    elif stat == "se":
                        headerDict[col].append(theSE[col])
                    elif stat == "n":
                        headerDict[col].append(theN[col])

            # print(headerDict)
            # for k,v in headerDict.items():
            #    print(k, v)

            # dict to pandas dataframe
            df = pd.DataFrame(headerDict).T
            df.to_excel(writer, sheet_name="summary")

            # set the column widths in excel sheet 'cardiac'
            columnWidth = 25
            worksheet = writer.sheets["summary"]  # pull worksheet object
            for idx, col in enumerate(df):  # loop through all columns
                worksheet.set_column(idx, idx, columnWidth)  # set column width

            #
            # 'params' sheet with all detection params
            # need to convert list values in dict to string (o.w. we get one row per item in list)
            exportDetectionDict = {}
            for k, v in dDict.items():
                # v is a dict from bDetection
                if isinstance(v, list):
                    v = f'"{v}"'
                exportDetectionDict[k] = v
            # print('  === "params" sheet exportDetectionDict:', exportDetectionDict)
            # df = pd.DataFrame(exportDetectionDict, index=[0]).T # index=[0] needed when dict has all scalar values
            detection_df = pd.DataFrame(exportDetectionDict).T
            detection_df.to_excel(writer, sheet_name="params")
            # worksheet is <class 'xlsxwriter.worksheet.Worksheet'>
            worksheet = writer.sheets["params"]  # pull worksheet object
            # set first 20 columns to columnWidth
            columnWidth = 18
            startCol = 0
            stopCol = 20  # xlswriter.worksheet does not care about the stop column
            worksheet.set_column(0, stopCol, columnWidth)  # set column width

            #
            # 'cardiac' sheet with human readable stat names
            cardiac_df.to_excel(writer, sheet_name="cardiac")

            # set the column widths in excel sheet 'cardiac'
            columnWidth = 20
            worksheet = writer.sheets["cardiac"]  # pull worksheet object
            for idx, col in enumerate(cardiac_df):  # loop through all columns
                worksheet.set_column(idx, idx, columnWidth)  # set column width

            #
            # mean spike clip
            theseClips, theseClips_x, meanClip = self.ba.getSpikeClips(
                theMin, theMax, sweepNumber=self.sweepNumber
            )
            try:
                first_X = theseClips_x[0]  # - theseClips_x[0][0]
                # if verbose: print('    bExport.saveReport() saving mean clip to sheet "Avg Spike" from', len(theseClips), 'clips')
                df = pd.DataFrame(meanClip, first_X)
                df.to_excel(writer, sheet_name="Avg Spike")
            except IndexError as e:
                logger.warning("Got bad spike clips. Usually happend when 1-2 spikes")

            writer.save()

        #
        # save a csv text file
        #
        analysisName = ""
        if alsoSaveTxt:
            # this also saves
            analysisName, df0 = self.getReportDf(theMin, theMax, savefile)

            #
            # save mean spike clip

            not_used_theseClips, theseClips_x, meanClip = self.ba.getSpikeClips(
                theMin, theMax, sweepNumber=self.sweepNumber
            )
            if len(theseClips_x) == 0:
                pass
            else:
                first_X = theseClips_x[0]  # - theseClips_x[0][0]
                first_X = np.array(first_X)
                first_X /= self.ba.fileLoader.dataPointsPerMs  # pnts to ms
                # if verbose: print('    bExport.saveReport() saving mean clip to sheet "Avg Spike" from', len(theseClips), 'clips')
                # dfClip = pd.DataFrame(meanClip, first_X)
                dfClip = pd.DataFrame.from_dict({"xMs": first_X, "yVm": meanClip})
                # load clip based on analysisname (with start/stop seconds)
                analysisname = df0["analysisname"].iloc[
                    0
                ]  # name with start/stop seconds
                logger.info(f"analysisname: {analysisname}")
                clipFileName = analysisname + "_clip.csv"
                tmpPath, tmpFile = os.path.split(savefile)
                tmpPath = os.path.join(tmpPath, "analysis")
                # dir is already created in getReportDf
                if not os.path.isdir(tmpPath):
                    os.mkdir(tmpPath)
                clipSavePath = os.path.join(tmpPath, clipFileName)
                logger.info(f"clipSavePath: {clipSavePath}")
                dfClip.to_csv(clipSavePath)
            #
            theRet = df0
        #
        return analysisName, theRet

    def getReportDf(self, theMin, theMax, savefile):
        """Get spikes as a Pandas DataFrame, one row per spike.

        Args:
            theMin (float): xxx
            theMax (float): xxx
            savefile (str): .xls file path

        Returns:
            df: Pandas DataFrame
        """
        filePath, fileName = os.path.split(os.path.abspath(savefile))

        dDict = self.ba.getDetectionDict()

        # make an analysis folder
        filePath = os.path.join(filePath, "analysis")
        if not os.path.isdir(filePath):
            logger.info(f"Making output folder: {filePath}")
            os.mkdir(filePath)

        textFileBaseName, tmpExtension = os.path.splitext(fileName)
        textFilePath = os.path.join(filePath, textFileBaseName + ".csv")

        # save header
        textFileHeader = OrderedDict()
        textFileHeader["file"] = self.ba.fileLoader.filename
        # textFileHeader['condition1'] = self.ba.condition1
        # textFileHeader['condition2'] = self.ba.condition2
        # textFileHeader['condition3'] = self.ba.condition3
        textFileHeader["cellType"] = dDict["cellType"]
        textFileHeader["sex"] = dDict["sex"]
        textFileHeader["condition"] = dDict["condition"]
        #
        textFileHeader["dateAnalyzed"] = self.ba.dateAnalyzed
        textFileHeader["detectionType"] = dDict["detectionType"]
        textFileHeader["dvdtThreshold"] = [dDict["dvdtThreshold"]]
        textFileHeader["mvThreshold"] = [dDict["mvThreshold"]]
        # textFileHeader['medianFilter'] = self.ba.medianFilter
        textFileHeader["startSeconds"] = "%.2f" % (theMin)
        textFileHeader["stopSeconds"] = "%.2f" % (theMax)
        # textFileHeader['startSeconds'] = self.ba.startSeconds
        # textFileHeader['stopSeconds'] = self.ba.stopSeconds
        textFileHeader["currentSweep"] = "Default 0"  # self.ba.currentSweep
        textFileHeader["numSweeps"] = self.ba.fileLoader.numSweeps
        # textFileHeader['theMin'] = theMin
        # textFileHeader['theMax'] = theMax

        # 20210125, this is not needed, we are saviing pandas df below ???
        headerStr = ""
        for k, v in textFileHeader.items():
            headerStr += k + "=" + str(v) + ";"
        headerStr += "\n"
        # print('headerStr:', headerStr)
        with open(textFilePath, "w") as f:
            f.write(headerStr)

        # df = self.report(theMin, theMax)
        df = self.ba.asDataFrame()

        # we need a column indicating (path), the original .abf file
        # along with (start,stop) which should make this analysis unique?
        minStr = "%.2f" % (theMin)
        maxStr = "%.2f" % (theMax)
        minStr = minStr.replace(".", "_")
        maxStr = maxStr.replace(".", "_")
        tmpPath, tmpFile = os.path.split(self.ba.fileLoader.filepath)
        tmpFile, tmpExt = os.path.splitext(tmpFile)
        analysisName = tmpFile + "_s" + minStr + "_s" + maxStr
        logger.info(f"minStr:{minStr} maxStr:{maxStr} analysisName:{analysisName}")
        df["analysisname"] = analysisName

        # should be filled in by self.ba.report
        # df['Condition'] =     df['condition1']
        # df['File Number'] =     df['condition2']
        # df['Sex'] =     df['condition3']
        # df['Region'] =     df['condition4']
        df["filename"] = [
            os.path.splitext(os.path.split(x)[1])[0] for x in df["file"].tolist()
        ]

        #
        logger.info("saving text file: {textFilePath}")
        # df.to_csv(textFilePath, sep=',', index_label='index', mode='a')
        df.to_csv(textFilePath, sep=",", index_label="index", mode="w")

        return analysisName, df


def test():
    path = "data/19114001.abf"
    ba = sanpy.bAnalysis(path)

    bd = sanpy.bDetection()  # gets default
    dDict = bd.getDetectionDict("SA Node")
    ba.spikeDetect(dDict)

    be = bExport(ba)
    df = be.getSummary()
    logger.info("")
    print(df)


if __name__ == "__main__":
    test()
